"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

This is an example extension that allow you to call an external service to retreive the energy consumption data
of the customer. Currently contains no authentication
"""
from programy.utils.logging.ylogger import YLogger
from programy.extensions.base import Extension


class SentimentExtension(Extension):

    def _check_enabled(self, client_context):
        if client_context.bot.sentiment_analyser is not None:
            return "SENTIMENT ENABLED"

        else:
            return "SENTIMENT DISABLED"

    def _calc_conversation_sentiment(self, client_context):
        conversation = client_context.bot.get_conversation(client_context)
        positivity, subjectivity = conversation.calculate_sentiment_score()
        if client_context.bot.sentiment_scores is not None:
            pos_str = client_context.bot.sentiment_scores.positivity(positivity)
            sub_str = client_context.bot.sentiment_scores.subjectivity(subjectivity)
            return "SENTIMENT FEELING %s AND %s" % (pos_str, sub_str)

        return "SENTIMENT FEELING NEUTRAL AND NEUTRAL"

    def _calc_question_sentiment(self, client_context, nth_question):
        conversation = client_context.bot.get_conversation(client_context)
        try:
            question = conversation.previous_nth_question(nth_question)

            positivity, subjectivity = question.calculate_sentinment_score()

            pos_str = client_context.bot.sentiment_scores.positivity(positivity)
            sub_str = client_context.bot.sentiment_scores.subjectivity(subjectivity)

            return "SENTIMENT FEELING %s AND %s" % (pos_str, sub_str)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to calculate sentiment", excep)
            return "SENTIMENT FEELING NEUTRAL AND NEUTRAL"

    def _calc_feeling(self, client_context, words):
        if client_context.bot.sentiment_analyser is not None:

            if len(words) >= 3:
                if words[2] == 'LAST':
                    if len(words) == 4:
                        if words[3].isdigit():
                            return self._calc_question_sentiment(client_context, int(words[3]))

                elif words[2] == 'OVERALL':
                    return self._calc_conversation_sentiment(client_context)

            return "SENTIMENT INVALID COMMAND"

        else:
            return "SENTIMENT DISABLED"

    def _current_score(self, client_context, words):

        conversation = client_context.bot.get_conversation(client_context)

        assert conversation is not None

        if len(words) == 3:

            if words[2] == 'NUMERIC':
                return "SENTIMENT SCORES POSITIVITY %s SUBJECTIVITY %s" % (conversation.properties['positivity'],
                                                                           conversation.properties['subjectivity'])

            if words[2] == 'TEXT':
                pos_str = client_context.bot.sentiment_scores.positivity(float(conversation.properties['positivity']),
                                                                         client_context)
                subj_str = client_context.bot.sentiment_scores.subjectivity(
                    float(conversation.properties['subjectivity']), client_context)
                return "SENTIMENT SCORES POSITIVITY %s SUBJECTIVITY %s" % (pos_str, subj_str)

        return "SENTIMENT INVALID COMMAND"

    def _calc_score(self, client_context, words):
        if client_context.bot.sentiment_analyser is not None:

            text = " ".join(words[2:])

            positivity, subjectivity = client_context.bot.sentiment_analyser.analyse_all(text)

            pos_str = "UNKNOWN"
            subj_str = "UNKNOWN"
            if client_context.bot.sentiment_scores is not None:
                pos_str = client_context.bot.sentiment_scores.positivity(positivity, client_context)
                subj_str = client_context.bot.sentiment_scores.subjectivity(subjectivity, client_context)

            return "SENTIMENT SCORES POSITIVITY %s SUBJECTIVITY %s" % (pos_str, subj_str)

        else:
            return "SENTIMENT DISABLED"

    def _get_positivity(self, client_context, words):
        positivity = float(words[2])
        return client_context.bot.sentiment_scores.positivity(positivity, client_context)

    def _get_subjectivity(self, client_context, words):
        subjectivity = float(words[2])
        return client_context.bot.sentiment_scores.subjectivity(subjectivity, client_context)

    # execute() is the interface that is called from the <extension> tag in the AIML
    def execute(self, client_context, data):
        YLogger.debug(client_context, "Sentiment - Calling external service for with extra data [%s]", data)

        # SENTIMENT SCORE <TEXT STRING>
        # SENTIMENT FEELING <TEXT STRING>
        # SENTIMENT ENABLED
        # SENTIMENT POSITIVITY <VALUE>
        # SENTIMENT SUBJECTIVITY <VALUE>

        words = data.split(" ")
        if words[0] == "SENTIMENT":

            if len(words) >= 2:
                if words[1] == "CURRENT":
                    return self._current_score(client_context, words)

                if words[1] == "SCORE":
                    return self._calc_score(client_context, words)

                if words[1] == "FEELING":
                    return self._calc_feeling(client_context, words)

                if words[1] == 'ENABLED':
                    return self._check_enabled(client_context)

                if words[1] == 'POSITIVITY':
                    return self._get_positivity(client_context, words)

                if words[1] == 'SUBJECTIVITY':
                    return self._get_subjectivity(client_context, words)

        return "SENTIMENT INVALID COMMAND"



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader
from programy.config.bot.splitter import BotSentenceSplitterConfiguration
from programy.activate import Activatable


class SentenceSplitter(Activatable):
    ALL_PUNCTUATION = re.compile(r'[:\'";,.?!\(\)\-"]')

    def __init__(self, splitter_config):
        Activatable.__init__(self)

        assert splitter_config is not None
        assert isinstance(splitter_config, BotSentenceSplitterConfiguration)

        self._configuration = splitter_config

    def split(self, text):
        raise NotImplementedError()  # pragma: no cover

    def remove_punctuation(self, text):
        return SentenceSplitter.ALL_PUNCTUATION.sub('', text)

    @staticmethod
    def initiate_sentence_splitter(splitter_config):
        if splitter_config.classname is not None:
            try:
                YLogger.info(None, "Loading sentence splitter from class [%s]", splitter_config.classname)
                splitter_class = ClassLoader.instantiate_class(splitter_config.classname)
                sentence_splitter = splitter_class(splitter_config)
                return sentence_splitter

            except Exception as excep:
                YLogger.exception(None, "Failed to initiate sentence splitter", excep)

        else:
            YLogger.warning(None, "No configuration setting for sentence splitter!")

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.parser.exceptions import ParserException
from programy.utils.text.text import TextUtils


class QueryBase:

    def __init__(self, subj, pred, obj):
        self._subj = subj
        self._pred = pred
        self._obj = obj

    @property
    def subj(self):
        return self._subj

    @property
    def pred(self):
        return self._pred

    @property
    def obj(self):
        return self._obj

    def to_xml(self, client_context):
        xml = "<subj>%s</subj>" % self._subj.resolve(client_context)
        xml += "<pred>%s</pred>" % self._pred.resolve(client_context)
        xml += "<obj>%s</obj>" % self._obj.resolve(client_context)
        return xml

    def execute(self, client_context, variables=None):
        del variables
        del client_context
        return []

    def get_rdf(self, client_context):
        subj = self.subj.resolve(client_context)
        if subj.startswith("?") is False:
            subj = subj.upper()
        pred = self.pred.resolve(client_context)
        if pred.startswith("?") is False:
            pred = pred.upper()
        obj = self.obj.resolve(client_context)
        return subj, pred, obj


class Query(QueryBase):

    def __init__(self, subj, pred, obj):
        QueryBase.__init__(self, subj, pred, obj)

    def get_xml_type(self):
        return "q"

    def to_xml(self, client_context):
        xml = "<q>"
        xml += super(Query, self).to_xml(client_context)
        xml += "</q>"
        return xml

    def _get_matched_tuples(self, client_context, subj, pred, obj):
        return client_context.brain.rdf.matched_as_tuples(subj, pred, obj)

    def execute(self, client_context, variables=None):
        subj, pred, obj = self.get_rdf(client_context)
        if variables is None:
            tuples = self._get_matched_tuples(client_context, subj, pred, obj)
            results = []
            for atuple in tuples:
                results.append([["subj", atuple[0]], ["pred", atuple[1]], ["obj", atuple[2]]])
            return results
        else:
            tuples = client_context.brain.rdf.match_to_vars(subj, pred, obj)
            return tuples


class NotQuery(QueryBase):

    def __init__(self, subj, pred, obj):
        QueryBase.__init__(self, subj, pred, obj)

    def get_xml_type(self):
        return "notq"

    def to_xml(self, client_context):
        xml = "<notq>"
        xml += super(NotQuery, self).to_xml(client_context)
        xml += "</notq>"
        return xml

    def _get_unmatched_tuples(self, client_context, subj, pred, obj):
        return client_context.brain.rdf.not_matched_as_tuples(subj, pred, obj)

    def execute(self, client_context, variables=None):
        subj, pred, obj = self.get_rdf(client_context)
        if variables is None:
            tuples = self._get_unmatched_tuples(client_context, subj, pred, obj)
            results = []
            for atuple in tuples:
                results.append([["subj", atuple[0]], ["pred", atuple[1]], ["obj", atuple[2]]])
            return results
        else:
            tuples = client_context.brain.rdf.not_match_to_vars(subj, pred, obj)
            return tuples


class TemplateSelectNode(TemplateNode):

    def __init__(self, queries=None, variables=None):
        TemplateNode.__init__(self)
        if queries is None:
            self._queries = []
        else:
            self._queries = queries[:]

        if variables is None:
            self._vars = []
        else:
            self._vars = variables[:]

    @property
    def queries(self):
        return self._queries

    @property
    def vars(self):
        return self._vars

    def encode_results(self, client_context, results):
        del client_context
        # At some point put a config item here that allows us to switch between
        # XML, JSON, Yaml, and Picke
        return json.dumps(results)

    def resolve_to_string(self, client_context):
        resolved = ""
        if self._queries:
            results = []

            for query in self._queries:
                query_results = query.execute(client_context, self.vars)
                results.append(query_results)

            if self.vars:
                results = client_context.brain.rdf.unify(self.vars, results)

            resolved = self.encode_results(client_context, results)

        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
        return resolved

    def to_string(self):
        return "[SELECT]"

    def to_xml(self, client_context):
        xml = "<select>"
        if self.vars:
            xml += "<vars>"
            xml += " ".join(self.vars)
            xml += "</vars>"
        if self._queries:
            for query in self._queries:
                xml += query.to_xml(client_context)
        xml += "</select>"
        return xml

    #######################################################################################################
    # SELECT_EXPRESSION ::== <person>TEMPLATE_EXPRESSION</person>

    def _parse_vars(self, variables):
        var_splits = variables.split(" ")
        for var_name in var_splits:
            self.vars.append(var_name)

    def _parse_query(self, graph, query_name, query):

        subj = None
        pred = None
        obj = None

        for child in query:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'subj':
                if child.text is not None and child.text.startswith("?"):
                    if child.text not in self.vars:
                        YLogger.debug(self, "Variable [%s] defined in query element [%s], but not in vars!",
                                      child.text, tag_name)
                        self.vars.append(child.text)

                subj = self.parse_children_as_word_node(graph, child)

            elif tag_name == 'pred':
                if child.text is not None and child.text.startswith("?"):
                    if child.text not in self.vars:
                        YLogger.debug(self, "Variable [%s] defined in query element [%s], but not in vars!",
                                      child.text, tag_name)
                        self.vars.append(child.text)

                pred = self.parse_children_as_word_node(graph, child)

            elif tag_name == 'obj':
                if child.text is not None and child.text.startswith("?"):
                    if child.text not in self.vars:
                        YLogger.debug(self, "Variable [%s] defined in query element [%s], but not in vars!",
                                      child.text, tag_name)
                        self.vars.append(child.text)

                obj = self.parse_children_as_word_node(graph, child)

            else:
                raise ParserException("Unknown query node [%s]" % tag_name)

        if subj is None:
            raise ParserException("<subj> element missing from select query")

        if pred is None:
            raise ParserException("<pred> element missing from select query")

        if obj is None:
            raise ParserException("<obj> element missing from select query")

        if query_name == "q":
            self._queries.append(Query(subj, pred, obj))
        else:
            self._queries.append(NotQuery(subj, pred, obj))

    def parse_expression(self, graph, expression):

        variables = expression.findall('./vars')
        if variables:
            if len(variables) > 1:
                YLogger.warning(self, "Multiple <vars> found in select tag, using first")

            self._parse_vars(variables[0].text)

        queries = expression.findall('./*')
        for query in queries:
            tag_name = TextUtils.tag_from_text(query.tag)
            if tag_name == 'q' or tag_name == 'notq':
                self._parse_query(graph, tag_name, query)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import logging
import traceback


class YLoggerSnapshot:  # pragma: no cover

    def __init__(self, criticals=0, fatals=0, errors=0, exceptions=0, warnings=0, infos=0, debugs=0):
        self._criticals = criticals
        self._fatals = fatals
        self._errors = errors
        self._exceptions = exceptions
        self._warnings = warnings
        self._infos = infos
        self._debugs = debugs

    def __str__(self):
        return "Critical(%d) Fatal(%d) Error(%d) Exception(%d) Warning(%d) Info(%d), Debug(%d)" % (
            self._criticals, self._fatals, self._errors, self._exceptions, self._warnings, self._infos, self._debugs
        )

    def to_json(self):
        return {
            "criticals": self._criticals,
            "fatals": self._fatals,
            "errors": self._errors,
            "exceptions": self._exceptions,
            "warnings": self._warnings,
            "infos": self._infos,
            "debugs": self._debugs
        }


class YLogger:  # pragma: no cover
    CRITICALS = 0
    FATALS = 0
    ERRORS = 0
    EXCEPTIONS = 0
    WARNINGS = 0
    INFOS = 0
    DEBUGS = 0

    @staticmethod
    def snapshot():
        return YLoggerSnapshot(YLogger.CRITICALS,
                               YLogger.FATALS,
                               YLogger.ERRORS,
                               YLogger.EXCEPTIONS,
                               YLogger.WARNINGS,
                               YLogger.INFOS,
                               YLogger.DEBUGS)

    @staticmethod
    def reset_snapshot():
        YLogger.CRITICALS = 0
        YLogger.FATALS = 0
        YLogger.ERRORS = 0
        YLogger.EXCEPTIONS = 0
        YLogger.WARNINGS = 0
        YLogger.INFOS = 0
        YLogger.DEBUGS = 0

    @staticmethod
    def format_message(caller, message):
        if caller is not None:
            if hasattr(caller, "ylogger_type"):
                log_type = caller.ylogger_type()
                if log_type == 'client':
                    return "[%s] - %s" % (caller.id, message)

                elif log_type == 'bot':
                    return "[%s] [%s] - %s" % (caller.client.id if caller.client is not None else "",
                                               caller.id, message)

                elif log_type == 'brain':
                    clientid = ""
                    botid = ""
                    if caller.bot is not None:
                        if caller.bot.client is not None:
                            clientid = caller.bot.client.id

                        botid = caller.bot.id

                    return "[%s] [%s] [%s] - %s" % (clientid, botid, caller.id, message)

                elif log_type == 'context':
                    return "[%s] [%s] [%s] [%s] - %s" % (caller.client.id if caller.client is not None else "",
                                                         caller.bot.id if caller.bot is not None else "",
                                                         caller.brain.id if caller.brain is not None else "",
                                                         caller.userid, message)
        return message

    @staticmethod
    def critical_enabled():
        return logging.getLogger().isEnabledFor(logging.CRITICAL)

    @staticmethod
    def fatal_enabled():
        return logging.getLogger().isEnabledFor(logging.FATAL)

    @staticmethod
    def error_enabled():
        return logging.getLogger().isEnabledFor(logging.ERROR)

    @staticmethod
    def warning_enabled():
        return logging.getLogger().isEnabledFor(logging.WARNING)

    @staticmethod
    def info_enabled():
        return logging.getLogger().isEnabledFor(logging.INFO)

    @staticmethod
    def debug_enabled():
        return logging.getLogger().isEnabledFor(logging.DEBUG)

    @staticmethod
    def critical(caller, message, *args, **kwargs):
        YLogger.CRITICALS += 1
        if YLogger.critical_enabled():
            logging.critical(YLogger.format_message(caller, message), *args, **kwargs)

    @staticmethod
    def fatal(caller, message, *args, **kwargs):
        YLogger.FATALS += 1
        if YLogger.fatal_enabled():
            logging.fatal(YLogger.format_message(caller, message), *args, **kwargs)

    @staticmethod
    def error(caller, message, *args, **kwargs):
        YLogger.ERRORS += 1
        if YLogger.error_enabled():
            logging.error(YLogger.format_message(caller, message), *args, **kwargs)

    @staticmethod
    def exception_nostack(caller, message, exception, *args, **kwargs):
        YLogger.EXCEPTIONS += 1
        if YLogger.error_enabled():
            excep_msg = "%s [%s]" % (message, str(exception))
            logging.error(YLogger.format_message(caller, excep_msg), *args, **kwargs)

    @staticmethod
    def exception(caller, message, exception, *args, **kwargs):
        YLogger.EXCEPTIONS += 1
        if YLogger.error_enabled():
            excep_msg = "%s [%s]" % (message, str(exception))
            logging.error(YLogger.format_message(caller, excep_msg), *args, **kwargs)
            tb_lines = [line.rstrip('\n') for line in
                        traceback.format_exception(exception.__class__, exception, exception.__traceback__)]
            for line in tb_lines:
                logging.exception(YLogger.format_message(caller, line))

    @staticmethod
    def warning(caller, message, *args, **kwargs):
        YLogger.WARNINGS += 1
        if YLogger.warning_enabled():
            logging.warning(YLogger.format_message(caller, message), *args, **kwargs)

    @staticmethod
    def info(caller, message, *args, **kwargs):
        YLogger.INFOS += 1
        if YLogger.info_enabled():
            logging.info(YLogger.format_message(caller, message), *args, **kwargs)

    @staticmethod
    def debug(caller, message, *args, **kwargs):
        YLogger.DEBUGS += 1
        if YLogger.debug_enabled():
            logging.debug(YLogger.format_message(caller, message), *args, **kwargs)

    @staticmethod
    def is_ylogger_method(method):
        return bool(method in [YLogger.critical, YLogger.fatal, YLogger.error, YLogger.exception,
                               YLogger.exception_nostack, YLogger.warning, YLogger.info, YLogger.debug])


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.config import ClientConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class XmppConfiguration(ClientConfigurationData):

    def __init__(self):
        ClientConfigurationData.__init__(self, "xmpp")
        self._server = None
        self._port = 5222
        self._xep_0030 = False
        self._xep_0004 = False
        self._xep_0060 = False
        self._xep_0199 = False

    @property
    def server(self):
        return self._server

    @property
    def port(self):
        return self._port

    @property
    def xep_0030(self):
        return self._xep_0030

    @property
    def xep_0004(self):
        return self._xep_0004

    @property
    def xep_0060(self):
        return self._xep_0060

    @property
    def xep_0199(self):
        return self._xep_0199

    def load_configuration_section(self, configuration_file, section, bot_root, subs: Substitutions = None):
        assert section is not None

        self._server = configuration_file.get_option(section, "server", subs=subs)
        self._port = configuration_file.get_int_option(section, "port", missing_value=5222, subs=subs)
        self._xep_0030 = configuration_file.get_bool_option(section, "xep_0030", subs=subs)
        self._xep_0004 = configuration_file.get_bool_option(section, "xep_0004", subs=subs)
        self._xep_0060 = configuration_file.get_bool_option(section, "xep_0060", subs=subs)
        self._xep_0199 = configuration_file.get_bool_option(section, "xep_0199", subs=subs)
        super(XmppConfiguration, self).load_configuration_section(configuration_file, section, bot_root, subs=subs)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['server'] = "talk.google.com"
            data['port'] = 5222
            data['xep_0030'] = True
            data['xep_0004'] = True
            data['xep_0060'] = True
            data['xep_0199'] = True
        else:
            data['server'] = self._server
            data['port'] = self._port
            data['xep_0030'] = self._xep_0030
            data['xep_0004'] = self._xep_0004
            data['xep_0060'] = self._xep_0060
            data['xep_0199'] = self._xep_0199

        super(XmppConfiguration, self).to_yaml(data, defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
import os
import os.path
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.property import PropertyStore
from programy.storage.stores.nosql.mongo.dao.property import Property
from programy.mappings.base import DoubleStringPatternSplitCollection
from programy.storage.entities.store import Store


class MongoPropertyStore(PropertyStore, MongoStore):
    PROPERTIES = 'properties'
    SPLIT_CHAR = ':'
    COMMENT = '#'
    NAME = 'name'
    VALUE = 'value'

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        PropertyStore.__init__(self)

    def collection_name(self):
        return MongoPropertyStore.PROPERTIES

    def empty_properties(self):
        self.empty()

    def add_property(self, name, value):
        collection = self.collection()
        prop = collection.find_one({MongoPropertyStore.NAME: name})
        if prop is not None:
            prop['value'] = value
            YLogger.info(self, "Replacing property [%s] = [%s]", name, value)
            result = collection.replace_one({'_id': prop['_id']}, prop)
            return bool(result.modified_count > 0)

        prop = Property(name, value)
        YLogger.info(self, "Adding property [%s] = [%s]", name, value)
        return self.add_document(prop)

    def add_properties(self, properties):
        for name, value in properties.items():
            self.add_property(name, value)

    def get_properties(self):
        collection = self.collection()
        props_collection = collection.find()
        properties = {}
        for prop in props_collection:
            properties[prop[MongoPropertyStore.NAME]] = prop[MongoPropertyStore.VALUE]
        return properties

    def load(self, collector, name=None):
        del name
        YLogger.info(self, "Loading properties from Mongo")
        self.load_all(collector)

    def load_all(self, collector):
        YLogger.info(self, "Loading all properties from Mongo")
        collector.empty()
        collection = self.collection()
        db_propertys = collection.find()
        for db_property in db_propertys:
            self.add_to_collection(collector, db_property[MongoPropertyStore.NAME],
                                   db_property[MongoPropertyStore.VALUE])

    def add_to_collection(self, collection, name, value):
        collection.add_property(name, value)

    def _process_line(self, line, verbose):
        line = line.strip()
        if line.startswith(MongoPropertyStore.COMMENT) is False:
            splits = line.split(MongoPropertyStore.SPLIT_CHAR)
            if len(splits) > 1:
                key = splits[0].strip()
                val = ":".join(splits[1:]).strip()
                if verbose is True:
                    YLogger.debug(self, "Adding %s property [%s=%s] to Mongo",
                                  self.collection_name(), key, val)
                return self.add_property(key, val)

        return False

    def _read_lines_from_file(self, filename, verbose):
        count = 0
        success = 0
        with open(filename, "r") as vars_file:
            for line in vars_file:
                if self._process_line(line, verbose) is True:
                    success += 1
                count += 1
        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        YLogger.info(self, "Uploading %s to Mongo from [%s]", filename, self.collection_name())

        try:
            return self._read_lines_from_file(filename, verbose)

        except Exception as excep:
            YLogger.exception(self, "Failed to upload %s from %s to Mongo", excep, self.collection_name(), filename)

        return 0, 0

    def split_into_fields(self, line):
        return DoubleStringPatternSplitCollection.split_line_by_pattern(line,
                                                                        DoubleStringPatternSplitCollection.
                                                                        RE_OF_SPLIT_PATTERN)


class MongoDefaultVariablesStore(MongoPropertyStore):
    DEFAULTS = 'defaults'

    def __init__(self, storage_engine):
        MongoPropertyStore.__init__(self, storage_engine)

    def collection_name(self):
        return MongoDefaultVariablesStore.DEFAULTS

    def add_defaults(self, defaults):
        self.add_properties(defaults)

    def get_default_values(self):
        return self.get_properties()

    def add_default(self, name, value):
        return self.add_property(name, value)


class MongoRegexesStore(MongoPropertyStore):
    REGEXES = 'regexes'

    def __init__(self, storage_engine):
        MongoPropertyStore.__init__(self, storage_engine)

    def collection_name(self):
        return MongoRegexesStore.REGEXES

    def add_regexes(self, regexes):
        self.add_properties(regexes)

    def get_regexes(self):
        return self.get_properties()

    def add_regex(self, name, regex):
        return self.add_property(name, regex)

    def add_to_collection(self, collection, name, value):
        try:
            collection.add_regex(name, re.compile(value, re.IGNORECASE))
        except Exception as excep:
            YLogger.exception(self, "Error adding regex to collection: [%s]", excep, value)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from programy.utils.logging.ylogger import YLogger

from programy.config.base import BaseConfigurationData
from programy.storage.stores.sql.config import SQLStorageConfiguration
from programy.storage.stores.file.config import FileStorageConfiguration
from programy.storage.stores.logger.config import LoggerStorageConfiguration
from programy.storage.stores.nosql.mongo.config import MongoStorageConfiguration
from programy.storage.stores.nosql.redis.config import RedisStorageConfiguration
from programy.storage.factory import StorageFactory
from programy.utils.substitutions.substitues import Substitutions


class StorageConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="storage")
        self._entity_store = {}
        self._store_configs = {}

    @property
    def entity_store(self):
        return self._entity_store

    @property
    def storage_configurations(self):
        return self._store_configs

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        storage = configuration_file.get_section(self._section_name, configuration)
        if storage is not None:

            entities = configuration_file.get_section("entities", storage)
            entity_types = configuration_file.get_child_section_keys("entities", storage)
            for entity in entity_types:
                entity_config = configuration_file.get_section(entity, entities)
                self._entity_store[entity] = entity_config

            stores = configuration_file.get_section("stores", storage)
            store_names = configuration_file.get_child_section_keys("stores", storage)
            for store in store_names:
                store_config = configuration_file.get_section(store, stores)
                keys = configuration_file.get_keys(store_config)

                if 'type' not in keys:
                    YLogger.error(None, "'type' section missing from client config stores element [%s], "
                                        "ignoring config", store)
                    continue

                if 'config' not in keys:
                    YLogger.error(None, "'config' section missing from client config stores element [%s], "
                                        "ignoring config", store)
                    continue

                storage_type = configuration_file.get_option(store_config, 'type', subs=subs)

                if storage_type == 'sql':
                    config = SQLStorageConfiguration()
                    config.load_config_section(configuration_file, store_config, bot_root, subs=subs)
                    self._store_configs[store] = config

                elif storage_type == 'mongo':
                    config = MongoStorageConfiguration()
                    config.load_config_section(configuration_file, store_config, bot_root, subs=subs)
                    self._store_configs[store] = config

                elif storage_type == 'redis':
                    config = RedisStorageConfiguration()
                    config.load_config_section(configuration_file, store_config, bot_root, subs=subs)
                    self._store_configs[store] = config

                elif storage_type == 'file':
                    config = FileStorageConfiguration()
                    config.load_config_section(configuration_file, store_config, bot_root, subs=subs)
                    self._store_configs[store] = config

                elif storage_type == 'logger':
                    config = LoggerStorageConfiguration()
                    config.load_config_section(configuration_file, store_config, bot_root, subs=subs)
                    self._store_configs[store] = config

                else:
                    YLogger.error(self, "Unknown storage configuration type [%s]", storage_type)

        else:
            YLogger.warning(self, "'storage' section missing from client config, using to defaults")

            self._entity_store = {}
            StorageConfiguration.add_default_entities(self._entity_store)

            self._store_configs = {}
            StorageConfiguration.add_default_stores(self._store_configs)

    def create_storage_config(self, file=True, sqlite=False, mongo=False, redis=False, logger=False):
        config = {}
        config['entities'] = {}
        StorageConfiguration.add_default_entities(config['entities'], file=file, sqlite=sqlite)

        config['stores'] = {}
        StorageConfiguration.add_default_stores(config['stores'], file=file, sqlite=sqlite, mongo=mongo, redis=redis, logger=logger)

        return config

    @staticmethod
    def add_default_stores(store_configs, file=True, sqlite=False, mongo=False, redis=False, logger=False):
        if sqlite is True:
            store_configs['sqlite'] = SQLStorageConfiguration()
        if mongo is True:
            store_configs['mongo'] = MongoStorageConfiguration()
        if redis is True:
            store_configs['redis'] = RedisStorageConfiguration()
        if file is True:
            store_configs['file'] = FileStorageConfiguration()
        if logger is True:
            store_configs['logger'] = LoggerStorageConfiguration()

    @staticmethod
    def add_default_stores_as_yaml(store_configs, file=True, sqlite=False, mongo=False, redis=False, logger=False):

        if file is True:
            store_configs['file'] = {}
            store_configs['file']['type'] = 'file'
            store_configs['file']['config'] = {}

            store = FileStorageConfiguration()
            store.to_yaml(store_configs['file']['config'], defaults=True)

        if sqlite is True:
            store_configs['sqlite'] = {}
            store_configs['sqlite']['type'] = 'sql'
            store_configs['sqlite']['config'] = {}

            store = SQLStorageConfiguration()
            store.to_yaml(store_configs['sqlite']['config'], defaults=True)

        if mongo is True:
            store_configs['mongo'] = {}
            store_configs['mongo']['type'] = 'mongo'
            store_configs['mongo']['config'] = {}

            store = MongoStorageConfiguration()
            store.to_yaml(store_configs['mongo']['config'], defaults=True)

        if redis is True:
            store_configs['redis'] = {}
            store_configs['redis']['type'] = 'redis'
            store_configs['redis']['config'] = {}

            store = RedisStorageConfiguration()
            store.to_yaml(store_configs['redis']['config'], defaults=True)

        if logger is True:
            store_configs['logger'] = {}
            store_configs['logger']['type'] = 'logger'
            store_configs['logger']['config'] = {}

            store = LoggerStorageConfiguration()
            store.to_yaml(store_configs['logger']['config'], defaults=True)

    @staticmethod
    def add_default_entities(entity_store, file=True, sqlite=False):

        if sqlite is True:
            entity_store[StorageFactory.USERS] = 'sqlite'
            entity_store[StorageFactory.LINKED_ACCOUNTS] = 'sqlite'
            entity_store[StorageFactory.LINKS] = 'sqlite'

        if file is True:
            entity_store[StorageFactory.CATEGORIES] = 'file'
            entity_store[StorageFactory.ERRORS] = 'file'
            entity_store[StorageFactory.DUPLICATES] = 'file'
            entity_store[StorageFactory.LEARNF] = 'file'

            entity_store[StorageFactory.CONVERSATIONS] = 'file'

            entity_store[StorageFactory.MAPS] = 'file'
            entity_store[StorageFactory.SETS] = 'file'
            entity_store[StorageFactory.RDF] = 'file'

            entity_store[StorageFactory.DENORMAL] = 'file'
            entity_store[StorageFactory.NORMAL] = 'file'
            entity_store[StorageFactory.GENDER] = 'file'
            entity_store[StorageFactory.PERSON] = 'file'
            entity_store[StorageFactory.PERSON2] = 'file'
            entity_store[StorageFactory.REGEX_TEMPLATES] = 'file'

            entity_store[StorageFactory.PROPERTIES] = 'file'
            entity_store[StorageFactory.DEFAULTS] = 'file'
            entity_store[StorageFactory.VARIABLES] = 'file'

            entity_store[StorageFactory.TWITTER] = 'file'

            entity_store[StorageFactory.SPELLING_CORPUS] = 'file'

            entity_store[StorageFactory.LICENSE_KEYS] = 'file'

            entity_store[StorageFactory.PATTERN_NODES] = 'file'
            entity_store[StorageFactory.TEMPLATE_NODES] = 'file'

            entity_store[StorageFactory.BINARIES] = 'file'
            entity_store[StorageFactory.BRAINTREE] = 'file'

            entity_store[StorageFactory.PREPROCESSORS] = 'file'
            entity_store[StorageFactory.POSTPROCESSORS] = 'file'
            entity_store[StorageFactory.POSTQUESTIONPROCESSORS] = 'file'

            entity_store[StorageFactory.USERGROUPS] = 'file'

            entity_store[StorageFactory.TRIGGERS] = 'file'

            entity_store[StorageFactory.OOBS] = 'file'

            entity_store[StorageFactory.SERVICES] = 'file'

    def to_yaml(self, data, defaults=True):

        data['entities'] = {}
        data['stores'] = {}

        if defaults is True:
            StorageConfiguration.add_default_entities(data['entities'])
            StorageConfiguration.add_default_stores_as_yaml(data['stores'])

        else:
            data['entities'] = {}
            for key, value in self._entity_store.items():
                data['entities'][key] = value

            for name, value in self._store_configs.items():
                data['stores'][name] = {}
                data['stores'][name] = value




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.conversation import ConversationStore
from programy.storage.stores.sql.dao.conversation import Conversation as ConversationDAO
from programy.storage.stores.sql.dao.conversation import Question as QuestionDAO
from programy.storage.stores.sql.dao.conversation import Sentence as SentenceDAO
from programy.storage.stores.sql.dao.conversation import ConversationProperty as ConversationPropertyDAO
from programy.storage.stores.sql.dao.conversation import Match as MatchDAO
from programy.storage.stores.sql.dao.conversation import MatchNode as MatchNodeDAO
from programy.dialog.question import Question
from programy.dialog.sentence import Sentence
from programy.parser.pattern.match import Match
from programy.parser.pattern.matchcontext import MatchContext


class SQLConversationStore(SQLStore, ConversationStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        ConversationStore.__init__(self)

    def _get_all(self):
        raise Exception("SQL Conversations uses complex multi table data structure, do not call _get_all()")

    def empty(self):
        self._storage_engine.session.query(ConversationDAO).delete()
        self._storage_engine.session.query(QuestionDAO).delete()
        self._storage_engine.session.query(SentenceDAO).delete()
        self._storage_engine.session.query(MatchDAO).delete()
        self._storage_engine.session.query(MatchNodeDAO).delete()
        self._storage_engine.session.query(ConversationPropertyDAO).delete()

    def _get_conversation_dao(self, client_context):
         return self._storage_engine.session.query(ConversationDAO). \
            filter(ConversationDAO.clientid == client_context.client.id,
                   ConversationDAO.userid == client_context.userid,
                   ConversationDAO.botid == client_context.bot.id,
                   ConversationDAO.brainid == client_context.brain.id).first()

    def store_conversation(self, client_context, conversation, commit=True):

        YLogger.debug(client_context, "Storing Conversation....")
        conversationdao = self._get_conversation_dao(client_context)
        if conversationdao is None:
            conversationdao = ConversationDAO(clientid=client_context.client.id,
                                              userid=client_context.userid,
                                              botid=client_context.bot.id,
                                              brainid=client_context.brain.id,
                                              maxhistories=conversation.max_histories)

            self._storage_engine.session.add(conversationdao)
            self._storage_engine.session.flush()
            YLogger.debug(client_context, "Wrote conversation %s", conversationdao)

        else:
            YLogger.debug(client_context, "Conversation already exists, %s", conversationdao)

        self._write_properties_to_db(client_context,
                                     conversationdao.id, 0,
                                     ConversationPropertyDAO.CONVERSATION,
                                     conversation.properties)

        self._write_questions_to_db(client_context, conversationdao.id, conversation)

        self.commit(commit)

    def _write_questions_to_db(self, client_context, conversationid, conversation):
        question_no = 1
        for question in conversation.questions:
            self._write_question_to_db(client_context, conversationid, question, question_no)
            question_no += 1

        self._storage_engine.session.flush()

    def _get_question_dao(self, conversationid, question_no):
        return self._storage_engine.session.query(QuestionDAO). \
            filter(QuestionDAO.conversationid == conversationid,
                   QuestionDAO.questionno == question_no).first()

    def _write_question_to_db(self, client_context, conversationid, question, question_no):

        questiondao = self._get_question_dao(conversationid, question_no)
        if questiondao is None:
            questiondao = QuestionDAO(
                conversationid=conversationid,
                questionno=question_no,
                srai=question.srai
            )

            self._storage_engine.session.add(questiondao)
            self._storage_engine.session.flush()
            YLogger.debug(client_context, "Writing question %s", questiondao)

        else:
            YLogger.debug(client_context, "Question already exists, %s", questiondao)

        self._write_properties_to_db(client_context,
                                     conversationid, questiondao.id,
                                     ConversationPropertyDAO.QUESTION,
                                     question.properties)

        self._write_sentences_to_db(client_context, questiondao.id, question)

    def _write_sentences_to_db(self, client_context, questionid, question):
        sentence_no = 1
        for sentence in question.sentences:
            self._write_sentence_to_db(client_context, questionid, sentence, sentence_no)
            sentence_no += 1
        self._storage_engine.session.flush()

    def _get_sentence_dao(self, questionid, sentence_no):
        return self._storage_engine.session.query(SentenceDAO). \
            filter(SentenceDAO.questionid == questionid,
                   SentenceDAO.sentenceno == sentence_no).first()

    def _write_sentence_to_db(self, client_context, questionid, sentence, sentence_no):

        sentencedao = self._get_sentence_dao(questionid, sentence_no)
        if sentencedao is None:
            sentencedao = SentenceDAO(questionid=questionid,
                                      sentenceno=sentence_no,
                                      sentence=sentence.text(client_context),
                                      response=sentence.response,
                                      subjectivity=sentence.subjectivity,
                                      positivity=sentence.positivity)

            self._storage_engine.session.add(sentencedao)
            self._storage_engine.session.flush()
            YLogger.debug(client_context, "Writing sentence %s", sentencedao)

        else:
            YLogger.debug(client_context, "Sentence already exists, %s", sentencedao)

        if sentence.matched_context is not None:
            self._write_match_context_to_db(client_context, sentencedao.id, sentence.matched_context)

    def _get_match_dao(self, sentenceid):
        return self._storage_engine.session.query(MatchDAO). \
            filter(MatchDAO.sentenceid == sentenceid).first()

    def _write_match_context_to_db(self, client_context, sentenceid, matched_context):

        matchdao = self._get_match_dao(sentenceid)
        if matchdao is None:
            matchdao = MatchDAO(sentenceid=sentenceid,
                                max_search_depth=matched_context.max_search_depth,
                                max_search_timeout=matched_context.max_search_timeout,
                                sentence=matched_context.sentence,
                                response=matched_context.response,
                                score=matched_context.calculate_match_score())

            self._storage_engine.session.add(matchdao)
            self._storage_engine.session.flush()
            YLogger.debug(client_context, "Writing match %s", matchdao)

        else:
            YLogger.debug(client_context, "Match already exists %s", matchdao)

        self._write_matches_to_db(client_context, matched_context, matchdao.id)

    def _get_matchnode_dao(self, matchid, match_count):
       return self._storage_engine.session.query(MatchNodeDAO). \
            filter(MatchNodeDAO.matchid == matchid,
                   MatchNodeDAO.matchcount == match_count).first()

    def _write_matches_to_db(self, client_context, matched_context, matchid):
        match_count = 1
        for match in matched_context.matched_nodes:
            matchnodedao = self._get_matchnode_dao(matchid, match_count)
            if matchnodedao is None:
                matchnodedao = MatchNodeDAO(matchid=matchid,
                                            matchcount=match_count,
                                            matchtype=Match.type_to_string(match.matched_node_type),
                                            matchnode=match.matched_node_str,
                                            matchstr=match.joined_words(client_context),
                                            wildcard=match.matched_node_wildcard,
                                            multiword=match.matched_node_multi_word)

                self._storage_engine.session.add(matchnodedao)
                YLogger.debug(client_context, "Writing matched node %s", matchnodedao)

            else:
                YLogger.debug(client_context, "Matched node already exists, %s", matchnodedao)

            match_count += 1

    def _get_property_dao(self, conversationid, questionid, proptype, name):
       return self._storage_engine.session.query(ConversationPropertyDAO). \
            filter(ConversationPropertyDAO.conversationid == conversationid,
                   ConversationPropertyDAO.questionid == questionid,
                   ConversationPropertyDAO.type == proptype,
                   ConversationPropertyDAO.name == name).first()

    def _write_properties_to_db(self, client_context, conversationid, questionid, proptype, properties):
        for name, value in properties.items():

            propertydao = self._get_property_dao(conversationid, questionid, proptype, name)
            if propertydao is None:
                propertydao = ConversationPropertyDAO(conversationid=conversationid,
                                                      questionid=questionid,
                                                      type=proptype,
                                                      name=name,
                                                      value=value)

                self._storage_engine.session.add(propertydao)
                self._storage_engine.session.flush()
                YLogger.debug(client_context, "Writing property %s", propertydao)

            elif propertydao.value != value:
                propertydao.value = value
                self._storage_engine.session.flush()
                YLogger.debug(client_context, "Updating property %s", propertydao)
            else:
                YLogger.debug(client_context, "Property already exists, %s", propertydao)

    def load_conversation(self, client_context, conversation):

        YLogger.debug(client_context, "Loading conversation")
        conversationdao = self._get_conversation_dao(client_context)
        if conversationdao is None:
            YLogger.debug(client_context, "No matching conversation in database")
            return False

        YLogger.debug(client_context, "Loaded conversation %s", conversationdao)

        conversation.max_histories = conversationdao.maxhistories

        self._read_properties_from_db(client_context,
                                      conversationdao.id, 0,
                                      ConversationPropertyDAO.CONVERSATION, conversation.properties)

        self._read_questions_from_db(client_context,
                                     conversationdao.id,
                                     conversation)

        return True

    def _read_questions_from_db(self, client_context, conversationid, conversation):

        questiondaos = self._storage_engine.session.query(QuestionDAO). \
            filter(QuestionDAO.conversationid == conversationid)

        for questiondao in questiondaos:
            YLogger.debug(client_context, "Loading question %s", questiondao)

            question = Question(questiondao.srai)

            self._read_properties_from_db(client_context, conversationid, questiondao.id,
                                          ConversationPropertyDAO.QUESTION,
                                          question.properties)

            self._read_sentences_from_db(client_context, questiondao.id, question)

            conversation.record_dialog(question)

    def _read_sentences_from_db(self, client_context, questiondid, question):
        sentencedaos = self._storage_engine.session.query(SentenceDAO). \
            filter(SentenceDAO.questionid == questiondid)

        for sentencedao in sentencedaos:
            YLogger.debug(client_context, "Loading sentence %s", sentencedao)

            sentence = Sentence(client_context, sentencedao.sentence)

            sentence.response = sentencedao.response
            sentence.positivity = float(sentencedao.positivity)
            sentence.subjectivity = float(sentencedao.subjectivity)
            sentence.matched_context = MatchContext(0, 0)

            question.sentences.append(sentence)

            self._read_match_context_from_db(client_context, sentencedao.id, sentence.matched_context)

    def _read_match_context_from_db(self, client_context, sentenceid, matched_context):
        matchdao = self._storage_engine.session.query(MatchDAO). \
            filter(MatchDAO.sentenceid == sentenceid).first()

        if matchdao is not None:
            matched_context.max_search_depth = matchdao.max_search_depth
            matched_context.max_search_timeout = matchdao.max_search_timeout
            matched_context.sentence = matchdao.sentence
            matched_context.response = matchdao.response

            YLogger.debug(client_context, "Loading match context %s", matchdao)
            self._read_matches_from_db(client_context, matched_context, matchdao.id)

    def _read_matches_from_db(self, client_context, matched_context, matchid):
        matchnodedaos = self._storage_engine.session.query(MatchNodeDAO). \
            filter(MatchNodeDAO.matchid == matchid)

        for matchnodedao in matchnodedaos:
            YLogger.debug(client_context, "Loading match node %s", matchnodedao)

            match = Match(None, None, None)
            match.matched_node_type = Match.string_to_type(matchnodedao.matchtype)
            match.matched_node_str = matchnodedao.matchnode
            match.matched_node_words = client_context.brain.tokenizer.texts_to_words(matchnodedao.matchstr)
            match.matched_node_multi_word = matchnodedao.multiword

            matched_context.matched_nodes.append(match)

    def _read_properties_from_db(self, client_context, conversationid, questionid, proptype, properties):

        propertydaos = self._storage_engine.session.query(ConversationPropertyDAO). \
            filter(ConversationPropertyDAO.conversationid == conversationid,
                   ConversationPropertyDAO.questionid == questionid,
                   ConversationPropertyDAO.type == proptype)

        for propertydao in propertydaos:
            YLogger.debug(client_context, "Loading conversation property %s", propertydao)
            properties[propertydao.name] = propertydao.value



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import argparse
from programy.storage.entities.store import Store
from programy.storage.stores.sql.engine import SQLStorageEngine
from programy.storage.stores.sql.config import SQLStorageConfiguration
from programy.storage.stores.sql.store.categories import SQLCategoryStore
from programy.storage.stores.sql.store.maps import SQLMapsStore
from programy.storage.stores.sql.store.sets import SQLSetsStore
from programy.storage.stores.sql.store.rdfs import SQLRDFsStore
from programy.storage.stores.sql.store.properties import SQLPropertyStore
from programy.storage.stores.sql.store.properties import SQLDefaultVariableStore
from programy.storage.stores.sql.store.properties import SQLRegexStore
from programy.storage.stores.sql.store.lookups import SQLDenormalStore
from programy.storage.stores.sql.store.lookups import SQLNormalStore
from programy.storage.stores.sql.store.lookups import SQLGenderStore
from programy.storage.stores.sql.store.lookups import SQLPersonStore
from programy.storage.stores.sql.store.lookups import SQLPerson2Store
from programy.storage.stores.sql.store.spelling import SQLSpellingStore
from programy.storage.stores.sql.store.licensekeys import SQLLicenseKeysStore
from programy.storage.stores.sql.store.usergroups import SQLUserGroupStore
from programy.storage.stores.sql.store.nodes import SQLTemplateNodesStore
from programy.storage.stores.sql.store.nodes import SQLPatternNodesStore
from programy.storage.stores.sql.store.processors import SQLPreProcessorsStore
from programy.storage.stores.sql.store.processors import SQLPostProcessorsStore
from programy.storage.stores.sql.store.processors import SQLPostQuestionProcessorsStore
from programy.storage.stores.sql.store.oobs import SQLOOBsStore
from programy.storage.stores.sql.store.triggers import SQLTriggersStore
from programy.utils.console.console import outputLog


class Uploader:

    @staticmethod
    def upload(storetype, url, filename, dirname, subdir, extension, create=False, drop_all=False, verbose=False):
        outputLog(None, "")
        config = SQLStorageConfiguration()
        config.url = url
        config.create_db = create
        config.drop_all_first = drop_all

        engine = SQLStorageEngine(config)
        engine.initialise()
        store = Uploader._get_store(storetype, engine)

        outputLog(None, "Emptying [%s]" % storetype)
        store.empty()

        outputLog(None, "Loading [%s]" % storetype)

        if filename is not None:
            count, success = store.upload_from_file(filename, commit=True, verbose=verbose)
            outputLog(None, "Lines processed [%d]" % count)
            outputLog(None, "Entities successful [%s]" % success)

        elif dirname is not None:
            count, success = store.upload_from_directory(dirname, fileformat=Store.TEXT_FORMAT, extension=extension,
                                                         subdir=subdir, commit=True, verbose=verbose)
            outputLog(None, "Lines processed [%d]" % count)
            outputLog(None, "Entities successful [%s]" % success)

        else:
            raise Exception("You must specify either --file or --dir")

        return count, success

    @staticmethod
    def _get_store(storetype, engine):
        if storetype == 'categories':
            return SQLCategoryStore(engine)
        if storetype == 'maps':
            return SQLMapsStore(engine)
        if storetype == 'sets':
            return SQLSetsStore(engine)
        if storetype == 'rdfs':
            return SQLRDFsStore(engine)
        if storetype == 'preprocessors':
            return SQLPreProcessorsStore(engine)
        if storetype == 'postprocessors':
            return SQLPostProcessorsStore(engine)
        if storetype == 'postquestionprocessors':
            return SQLPostQuestionProcessorsStore(engine)
        if storetype == 'templatenodes':
            return SQLTemplateNodesStore(engine)
        if storetype == 'patternnodes':
            return SQLPatternNodesStore(engine)
        if storetype == 'properties':
            return SQLPropertyStore(engine)
        if storetype == 'defaults':
            return SQLDefaultVariableStore(engine)
        if storetype == 'regexes':
            return SQLRegexStore(engine)
        if storetype == 'denormals':
            return SQLDenormalStore(engine)
        if storetype == 'normals':
            return SQLNormalStore(engine)
        if storetype == 'genders':
            return SQLGenderStore(engine)
        if storetype == 'persons':
            return SQLPersonStore(engine)
        if storetype == 'person2s':
            return SQLPerson2Store(engine)
        if storetype == 'spelling':
            return SQLSpellingStore(engine)
        if storetype == 'licenses':
            return SQLLicenseKeysStore(engine)
        if storetype == 'usergroups':
            return SQLUserGroupStore(engine)
        if storetype == 'triggers':
            return SQLTriggersStore(engine)
        if storetype == 'oobs':
            return SQLOOBsStore(engine)

        raise Exception("Unknown entity storetype [%s]" % storetype)

    @staticmethod
    def create_arguments():
        loader_args = argparse.ArgumentParser(description='Program-Y Set SQL Loader')

        loader_args.add_argument('-e', '--entity', help="Name of entity to load")
        loader_args.add_argument('-u', '--url', help="SQL Alchemy connection url")
        loader_args.add_argument('-f', '--file', help="File to load")
        loader_args.add_argument('-d', '--dir', help="Directory to load files from")
        loader_args.add_argument('-x', '--extension', help="Extension of file to load (dir only)")
        loader_args.add_argument('-s', '--subdir', action='store_true', help="Recurse into all subdirectories")
        loader_args.add_argument('-c', '--create', action='store_true', help="Create database")
        loader_args.add_argument('-a', '--drop', action='store_true', help="Drop all table first")
        loader_args.add_argument('-v', '--verbose', action='store_true', help="Verbose output of loading process")

        return loader_args

    @staticmethod
    def get_args(arguments):
        return arguments.parse_args()       #pragma: no cover

    @staticmethod
    def run():
        arguments = Uploader.create_arguments()
        try:
            args = Uploader.get_args(arguments)
            Uploader.upload(args.entity,
                            args.url,
                            args.file,
                            args.dir,
                            args.subdir,
                            args.extension,
                            create=args.create,
                            drop_all=args.drop,
                            verbose=args.verbose)

        except Exception as excep:
            outputLog(None, "SQL Loader error occured - %s" % excep)
            arguments.print_help()
        except SystemExit as excep:
            outputLog(None, "SQL Loader error occured - %s" % excep)
            arguments.print_help()

if __name__ == '__main__':

    Uploader.run()                          #pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.utils.logging.ylogger import YLogger


######################################################################################################################
#
class TemplateNode:

    def __init__(self):
        self._children = []

    @property
    def children(self):
        return self._children

    def append(self, child):
        self._children.append(child)

    def dump(self, tabs, output_func, eol, verbose):
        self.output(tabs, output_func, eol, verbose)

    def output(self, tabs, output_func, eol, verbose):
        del verbose
        self.output_child(self, tabs, eol, output_func)

    def output_child(self, node, tabs, eol, output_func):
        for child in node.children:
            if output_func == print:    #pylint: disable=comparison-with-callable
                output_func("{0}{1}{2}".format(tabs, child.to_string(), eol))
            else:
                output_func(self, "{0}{1}{2}".format(tabs, child.to_string(), eol))
            self.output_child(child, tabs + "\t", eol, output_func)

    def resolve_children_to_string(self, client_context):
        words = []
        for child in self._children:
            words.append(child.resolve(client_context))

        return client_context.brain.tokenizer.words_to_texts(words)

    def resolve_to_string(self, client_context):
        return self.resolve_children_to_string(client_context)

    def resolve(self, client_context):
        try:
            resolved = self.resolve_to_string(client_context)
            YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
            return resolved
        except Exception as excep:
            YLogger.exception(client_context, "Template node failed to resolve", excep)
            return ""

    def to_string(self):
        return "[NODE]"

    def to_xml(self, client_context):
        return self.children_to_xml(client_context)

    def xml_tree(self, client_context):
        xml = "<template>"
        value = self.children_to_xml(client_context)
        xml += value
        xml += "</template>"
        return ET.fromstring(xml)

    def children_to_xml(self, client_context):
        xml = ""
        first = True
        for child in self.children:
            if first is not True:
                xml += " "
            first = False
            xml += child.to_xml(client_context)
        return xml

    def parse_text(self, graph, text):
        if text is not None:
            string = text.strip()
            if string:
                words = graph.aiml_parser.brain.tokenizer.texts_to_words(string)

                for word in words:
                    if word is not None and word:
                        word_class = graph.get_node_class_by_name('word')
                        word_node = word_class(word.strip())
                        self.children.append(word_node)
                return True
        return False

    def get_text_from_element(self, element):
        text = element.text
        if text is not None:
            text = text.strip()
            return text
        return None

    def get_tail_from_element(self, element):
        text = element.tail
        if text is not None:
            text = text.strip()
            if text == "":
                return None
            return text
        return None

    def parse_template_node(self, graph, pattern):

        head_text = self.get_text_from_element(pattern)
        head_result = self.parse_text(graph, head_text)

        found_sub = False
        for sub_pattern in pattern:
            graph.parse_tag_expression(sub_pattern, self)

            tail_text = self.get_tail_from_element(sub_pattern)
            self.parse_text(graph, tail_text)

            found_sub = True

        if head_result is False and found_sub is False:
            if hasattr(pattern, '_end_line_number'):  # pragma: no cover
                YLogger.warning(self, "No context in template tag at [line(%d), column(%d)]",  # pragma: no cover
                                pattern._end_line_number,  # pylint: disable=protected-access
                                pattern._end_column_number)  # pylint: disable=protected-access
            else:
                YLogger.warning(self, "No context in template tag")

    #######################################################################################################

    def add_default_star(self):
        return False

    def _parse_node(self, graph, expression):
        expression_text = self.parse_text(graph, self.get_text_from_element(expression))

        expression_children = False
        for child in expression:
            graph.parse_tag_expression(child, self)
            self.parse_text(graph, self.get_tail_from_element(child))
            expression_children = True

        if expression_text is False and expression_children is False:
            if self.add_default_star() is True:
                YLogger.debug(self, "Node has no content (text or children), default to <star/>")
                star_class = graph.get_node_class_by_name('star')
                star_node = star_class()
                self.append(star_node)

    #######################################################################################################

    def parse_attrib_value_as_word_node(self, graph, expression, attrib_name):
        node = graph.get_base_node()
        name_node = graph.get_word_node(expression.attrib[attrib_name])
        node.append(name_node)
        return node

    def parse_children_as_word_node(self, graph, child):
        node = graph.get_base_node()
        node.parse_text(graph, self.get_text_from_element(child))
        for sub_child in child:
            graph.parse_tag_expression(sub_child, node)
            node.parse_text(graph, self.get_text_from_element(child))
        return node

    def parse_expression(self, graph, expression):
        raise NotImplementedError("Never call this directly, call the subclass instead!")  # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.dialog.question import Question
from programy.dialog.sentence import Sentence
from programy.dialog.convo_mgr import ConversationManager
from programy.spelling.base import SpellingChecker
from programy.dialog.splitter.splitter import SentenceSplitter
from programy.dialog.joiner.joiner import SentenceJoiner
from programy.nlp.translate.base import BaseTranslator
from programy.nlp.sentiment.base import BaseSentimentAnalyser
from programy.nlp.sentiment.scores import SentimentScores
from programy.triggers.system import SystemTriggers
from programy.brainfactory import BrainFactory
from programy.clients.config import ClientConfigurationData
from programy.context import ClientContext
from programy.dialog.conversation import Conversation


class Bot:

    def __init__(self, config: ClientConfigurationData, client):

        assert config is not None
        assert client is not None

        self._questions = 0

        self._configuration = config
        self._client = client

        self._brain_factory = BrainFactory(self)

        self._question_depth = 0
        self._question_start_time = None

        self._spell_checker = None
        self.initiate_spellchecker()

        self._sentence_splitter = None
        self.initiate_sentence_splitter()

        self._sentence_joiner = None
        self.initiate_sentence_joiner()

        self._from_translator = None
        self._to_translator = None
        self.initiate_translator()

        self._sentiment_analyser = None
        self._sentiment_scores = None
        self.initiate_sentiment_analyser()

        self._conversation_mgr = ConversationManager(config.conversations)
        self._conversation_mgr.initialise(self._client.storage_factory)

    @staticmethod
    def ylogger_type() -> str:
        return "bot"

    @property
    def num_questions(self) -> int:
        return self._questions

    @property
    def id(self) -> str:
        return self._configuration.section_name

    @property
    def client(self):
        return self._client

    @property
    def configuration(self) -> ClientConfigurationData:
        return self._configuration

    @property
    def brain_factory(self) -> BrainFactory:
        return self._brain_factory

    @property
    def spell_checker(self) -> SpellingChecker:
        return self._spell_checker

    def get_question_counts(self) -> int:
        return self._brain_factory.get_question_counts()

    def initiate_spellchecker(self):
        if self.configuration.spelling is not None:
            self._spell_checker = SpellingChecker.initiate_spellchecker(self.configuration.spelling,
                                                                        self.client.storage_factory)

    @property
    def sentence_splitter(self):
        return self._sentence_splitter

    def initiate_sentence_splitter(self) -> SentenceSplitter:
        if self.configuration.splitter is not None:
            self._sentence_splitter = SentenceSplitter.initiate_sentence_splitter(self.configuration.splitter)

    @property
    def sentence_joiner(self) -> SentenceJoiner:
        return self._sentence_joiner

    def initiate_sentence_joiner(self):
        if self.configuration.joiner is not None:
            self._sentence_joiner = SentenceJoiner.initiate_sentence_joiner(self.configuration.joiner)

    @property
    def from_translator(self) -> BaseTranslator:
        return self._from_translator

    @property
    def to_translator(self) -> BaseTranslator:
        return self._to_translator

    def initiate_translator(self):
        if self.configuration.from_translator is not None:
            self._from_translator = BaseTranslator.initiate_translator(self.configuration.from_translator)

        if self.configuration.to_translator is not None:
            self._to_translator = BaseTranslator.initiate_translator(self.configuration.to_translator)

    @property
    def sentiment_analyser(self) -> BaseSentimentAnalyser:
        return self._sentiment_analyser

    @property
    def sentiment_scores(self) -> SentimentScores:
        return self._sentiment_scores

    def initiate_sentiment_analyser(self):
        if self.configuration.sentiment_analyser is not None:
            self._sentiment_analyser, self._sentiment_scores = BaseSentimentAnalyser.initiate_sentiment_analyser(
                self.configuration.sentiment_analyser)

    @property
    def brain(self):
        return self._brain_factory.select_brain()

    @property
    def conversations(self) -> ConversationManager:
        return self._conversation_mgr

    @property
    def default_response(self) -> str:
        return self.configuration.default_response

    @property
    def default_response_srai(self) -> str:
        return self.configuration.default_response_srai

    @property
    def exit_response(self) -> str:
        return self.configuration.exit_response

    @property
    def exit_response_srai(self) -> str:
        return self.configuration.exit_response_srai

    @property
    def initial_question(self) -> str:
        return self.configuration.initial_question

    @property
    def initial_question_srai(self) -> str:
        return self.configuration.initial_question_srai

    @property
    def override_properties(self) -> bool:
        return self.configuration.override_properties

    def get_version_string(self, client_context) -> str:

        assert client_context is not None

        if client_context.brain.properties.has_property("version"):
            # The old version of returning the version string, did not distinquish
            # between App and Grammar version
            return "%s, v%s, initiated %s" % (
                client_context.brain.properties.property("name"),
                client_context.brain.properties.property("version"),
                client_context.brain.properties.property("birthdate"))
        else:
            # This version now does
            return "%s, App: v%s Grammar v%s, initiated %s" % (
                client_context.brain.properties.property("name"),
                client_context.brain.properties.property("app_version"),
                client_context.brain.properties.property("grammar_version"),
                client_context.brain.properties.property("birthdate"))

    def has_conversation(self, client_context) -> bool:

        assert self._conversation_mgr is not None

        return self._conversation_mgr.has_conversation(client_context)

    def get_conversation(self, client_context: ClientContext) -> Conversation:

        assert self._conversation_mgr is not None

        return self._conversation_mgr.get_conversation(client_context)

    def check_spelling_before(self, client_context: ClientContext, each_sentence):
        if self.spell_checker is not None:
            self.spell_checker.check_spelling_before(client_context, each_sentence)

    def check_spelling_and_retry(self, client_context, each_sentence):
        if self.spell_checker is not None:
            return self.spell_checker.check_spelling_and_retry(client_context, each_sentence)
        return None

    def get_default_response(self, client_context):

        assert client_context is not None

        if self.default_response_srai is not None:
            sentence = Sentence(client_context, self.default_response_srai)
            default_response = client_context.brain.ask_question(client_context, sentence)
            if default_response is None or not default_response:
                default_response = self.default_response
            return default_response
        else:
            return self.default_response

    def get_initial_question(self, client_context):

        assert client_context is not None

        if self.initial_question_srai is not None:
            sentence = Sentence(client_context, self.initial_question_srai)
            initial_question = client_context.brain.ask_question(client_context, sentence)
            if initial_question is None or not initial_question:
                initial_question = self.initial_question
            return initial_question
        else:
            return self.initial_question

    def get_exit_response(self, client_context):

        assert client_context is not None

        if self.exit_response_srai is not None:
            sentence = Sentence(client_context, self.exit_response_srai)
            exit_response = client_context.brain.ask_question(client_context, sentence)
            if exit_response is None or not exit_response:
                exit_response = self.exit_response

            return exit_response

        else:
            return self.exit_response

    def _brain_pre_process_question(self, client_context, text):
        pre_processed = client_context.brain.pre_process_question(client_context, text)
        YLogger.debug(client_context, "Pre Processed (%s): %s", client_context.userid, pre_processed)
        return pre_processed

    def pre_process_text(self, client_context, text, srai):

        assert client_context is not None
        assert client_context.brain is not None

        if srai is False:
            pre_processed = self._brain_pre_process_question(client_context, text)
        else:
            pre_processed = text

        if pre_processed is None or pre_processed == "":
            assert self.configuration is not None
            pre_processed = self.configuration.empty_string

        return pre_processed

    def get_question(self, client_context, pre_processed, srai):
        if srai is False:
            return Question.create_from_text(client_context, pre_processed, srai=srai)
        else:
            return Question.create_from_text(client_context, pre_processed, split=False, srai=srai)

    def combine_answers(self, answers, srai):

        assert answers is not None
        assert self._sentence_joiner is not None

        return self._sentence_joiner.combine_answers(answers, srai)

    def post_process_response(self, client_context, response, srai):
        if srai is False:

            assert client_context is not None

            answer = client_context.brain.post_process_response(client_context, response).strip()
            if not answer:
                answer = self.get_default_response(client_context)

        else:
            answer = response

        return answer

    def _brain_post_process_question(self, client_context, question):
        return client_context.brain.post_process_question(client_context, question)

    def post_process_question(self, client_context, question, srai):
        if srai is False:
            assert client_context is not None
            answer = self._brain_post_process_question(client_context, question)
            if not answer:
                answer = self.get_default_response(client_context)
        else:
            answer = question

        return answer

    def log_answer(self, client_context, text, answer, responselogger):
        YLogger.debug(client_context, "Processed Response (%s): %s", client_context.userid, answer)

        if responselogger is not None:
            responselogger.log_response(text, answer)

    def ask_question(self, client_context, text, srai=False, responselogger=None):

        assert client_context is not None

        if srai is False:
            client_context.bot = self
            client_context.brain = client_context.bot.brain

        assert client_context.bot is not None
        assert client_context.brain is not None

        self._questions += 1

        client_context.mark_question_start(text)

        pre_processed = self.pre_process_text(client_context, text, srai=False)

        question = self.get_question(client_context, pre_processed, srai)

        conversation = self.get_conversation(client_context)
        if len(conversation.questions) == 0:
            if self.client.trigger_manager is not None:
                self.client.trigger_manager.trigger(SystemTriggers.CONVERSATION_START, client_context)

        assert conversation is not None

        conversation.record_dialog(question)

        answers = self.process_sentences(client_context, question, srai, responselogger)

        client_context.reset_question()

        if srai is True:
            conversation.pop_dialog()

        conversation.save_sentiment()

        self._conversation_mgr.save_conversation(client_context)

        if self.client.trigger_manager is not None and srai is False:
            self.client.trigger_manager.trigger(SystemTriggers.QUESTION_ASKED, client_context)

        return self.combine_answers(answers, srai)

    def process_sentences(self, client_context, question, srai, responselogger):
        answers = []
        sentence_no = 0
        for sentence in question.sentences:
            question.current_sentence_no = sentence_no
            answer = self.process_sentence(client_context, sentence, srai, responselogger)
            answers.append(answer)
            sentence_no += 1

        return answers

    def process_sentence(self, client_context, sentence, srai, responselogger):

        assert client_context is not None
        assert client_context.brain is not None

        client_context.check_max_recursion()
        client_context.check_max_timeout()

        if srai is False:
            self.check_spelling_before(client_context, sentence)

        response = client_context.brain.ask_question(client_context, sentence, srai)

        if response is None and srai is False:
            response = self.check_spelling_and_retry(client_context, sentence)

        if response is not None:
            return self.handle_response(client_context, sentence, response, srai, responselogger)
        else:
            return self.handle_none_response(client_context, sentence, responselogger)

    def handle_response(self, client_context, sentence, response, srai, responselogger):

        assert sentence is not None

        YLogger.debug(client_context, "Raw Response (%s): %s", client_context.userid, response)
        sentence.response = response

        sentence.calculate_sentinment_score(client_context)
        answer = self.post_process_response(client_context, response, srai)
        self.log_answer(client_context, sentence.text, answer, responselogger)

        return answer

    def handle_none_response(self, client_context, sentence, responselogger):

        assert sentence is not None

        sentence.response = self.post_process_question(client_context, sentence.text(client_context), False)

        sentence.calculate_sentinment_score(client_context)

        if responselogger is not None:
            responselogger.log_unknown_response(sentence)

        return sentence.response



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import List
from typing import NewType

import datetime
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.match import Match
from programy.parser.pattern.nodes.template import PatternTemplateNode

dt = NewType('dt', datetime.datetime)


class MatchContext:

    def __init__(self,
                 max_search_depth: int,
                 max_search_timeout: int,
                 matched_nodes: List = None,
                 template_node: PatternTemplateNode = None,
                 sentence: str = None,
                 response: str = None):
        self._max_search_depth = max_search_depth
        self._max_search_timeout = max_search_timeout
        self._total_search_start = datetime.datetime.now()
        self._matched_nodes = []
        if matched_nodes is not None:
            self._matched_nodes = matched_nodes.copy()
        if template_node is not None:
            assert isinstance(template_node, PatternTemplateNode)
        self._template_node = template_node
        self._sentence = sentence
        self._response = response

    @property
    def matched_nodes(self) -> List:
        return self._matched_nodes

    def set_matched_nodes(self, nodes: List):
        self._matched_nodes = nodes[:]

    @property
    def template_node(self) -> PatternTemplateNode:
        return self._template_node

    @template_node.setter
    def template_node(self, template_node: PatternTemplateNode):
        assert isinstance(template_node, PatternTemplateNode)
        self._template_node = template_node

    @property
    def response(self) -> str:
        return self._response

    @response.setter
    def response(self, response: str):
        self._response = response

    @property
    def sentence(self) -> str:
        return self._sentence

    @sentence.setter
    def sentence(self, sentence: str):
        self._sentence = sentence

    @property
    def max_search_depth(self) -> int:
        return self._max_search_depth

    @max_search_depth.setter
    def max_search_depth(self, depth: int):
        self._max_search_depth = depth

    @property
    def total_search_start(self) -> dt:
        return self._total_search_start

    @total_search_start.setter
    def total_search_start(self, start: dt):
        self._total_search_start = start

    @property
    def max_search_timeout(self) -> dt:
        return self._max_search_timeout

    @max_search_timeout.setter
    def max_search_timeout(self, timeout: dt):
        self._max_search_timeout = timeout

    def search_depth_exceeded(self, depth: int) -> bool:
        if self._max_search_depth == -1:
            return False

        return bool(depth > self._max_search_depth)

    def total_search_time(self) -> int:
        delta = datetime.datetime.now() - self._total_search_start
        return int(abs(delta.total_seconds()))

    def search_time_exceeded(self) -> bool:
        if self._max_search_timeout == -1:
            return False

        return bool(self.total_search_time() >= self._max_search_timeout)

    def add_match(self, match):
        self._matched_nodes.append(match)

    def pop_match(self):
        if self._matched_nodes:
            self._matched_nodes.pop()

    def pop_matches(self, matches_add):
        for _ in range(0, matches_add):
            self.pop_match()

    def matched(self):
        return bool(self._template_node is not None or self._response is not None)

    def _get_indexed_match_by_type(self, client_context, index, match_type):
        count = 1
        for matched_node in self._matched_nodes:
            if matched_node.matched_node_type == match_type and matched_node.matched_node_multi_word is True:
                if count == index:
                    return matched_node.joined_words(client_context)
                count += 1
        return None

    def star(self, client_context, index):
        return self._get_indexed_match_by_type(client_context, index, Match.WORD)

    def topicstar(self, client_context, index):
        return self._get_indexed_match_by_type(client_context, index, Match.TOPIC)

    def thatstar(self, client_context, index):
        return self._get_indexed_match_by_type(client_context, index, Match.THAT)

    def list_matches(self, client_context, output_func=YLogger.debug, tabs="\t", include_template=True):
        output_func(client_context, "%sMatches..." % tabs)
        count = 1
        if self._sentence is not None:
            output_func(client_context, "%sAsked: %s" % (tabs, self._sentence))
        for match in self._matched_nodes:
            output_func(client_context, "%s\t%d: %s" % (tabs, count, match.to_string(client_context)))
            count += 1
        output_func(client_context, "%sMatch score %.2f" % (tabs, self.calculate_match_score()))
        if include_template is True:
            if self.matched() is True:
                if self._response is not None:
                    output_func(client_context, "%s\tResponse: %s" % (tabs, self._response))
            else:
                output_func(client_context, "%s\tResponse: None" % tabs)

    def calculate_match_score(self):
        wildcards = 0
        words = 0
        for match in self._matched_nodes:
            if match.matched_node_type == Match.WORD:
                if match.matched_node_wildcard:
                    wildcards += 1
                else:
                    words += 1
        total = wildcards + words
        if total > 0:
            return (words // (wildcards + words)) * 100.00
        return 0.00

    def to_json(self):
        context={ "max_search_depth":self._max_search_depth,
                  "max_search_timeout": self._max_search_timeout,
                  "total_search_start": self._total_search_start.strftime("%d/%m/%Y, %H:%M:%S"),
                  "sentence": self._sentence,
                  "response": self._response,
                  "matched_nodes": []
                }

        for match in self._matched_nodes:
            context["matched_nodes"].append(match.to_json())

        return context

    @staticmethod
    def from_json(json_data):
        match_context = MatchContext(0, 0)

        match_context.max_search_depth = json_data["max_search_depth"]
        match_context.max_search_timeout = json_data["max_search_timeout"]
        match_context.total_search_start = datetime.datetime.strptime(json_data["total_search_start"],
                                                                      "%d/%m/%Y, %H:%M:%S")
        match_context.sentence = json_data["sentence"]
        match_context.response = json_data["response"]

        for match_data in json_data["matched_nodes"]:
            match_context.matched_nodes.append(Match.from_json(match_data))

        return match_context



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.parser.template.nodes.attrib import TemplateAttribNode
from programy.parser.template.nodes.word import TemplateWordNode
from programy.parser.template.nodes.word import TemplateNode
from programy.utils.text.text import TextUtils
from programy.utils.logging.ylogger import YLogger
from programy.parser.exceptions import ParserException


class TemplateXMLNode(TemplateAttribNode):

    def __init__(self):
        TemplateAttribNode.__init__(self)
        self._name = None
        self._attribs = {}

    @property
    def attribs(self):
        return self._attribs

    def set_attrib(self, attrib_name, attrib_value):
        self._attribs[attrib_name] = attrib_value

    def resolve_to_string(self, client_context):
        return self.to_xml(client_context)

    def to_string(self):
        return "[XML]"

    def to_xml(self, client_context):
        if self._name is not None:
            xml = "<%s" % self._name
            for attrib_name in self._attribs:
                attrib_node = self._attribs[attrib_name]
                attrib_value = attrib_node.to_xml(client_context)
                escaped = TextUtils.html_escape(attrib_value)
                escaped = escaped.replace(" ", "")
                xml += ' %s="%s"' % (attrib_name, escaped)

            xml += ">"
            child_xml = self.children_to_xml(client_context)
            xml += child_xml
            xml += "</%s>" % self._name
            return xml

        return ""

    def resolve(self, client_context):
        if self._name is not None:
            xml = "<%s" % self._name
            for attrib_name in self._attribs:
                attrib_node = self._attribs[attrib_name]
                attrib_value = attrib_node.resolve(client_context)
                escaped = TextUtils.html_escape(attrib_value)
                escaped = escaped.replace(" ", "")
                xml += ' %s="%s"' % (attrib_name, escaped)

            xml += ">"
            child_xml = self.resolve_children_to_string(client_context)
            xml += child_xml
            xml += "</%s>" % self._name
            return xml

        return ""

    def parse_expression(self, graph, expression):
        self._parse_node_with_xml_attribs(graph, expression)

    def _parse_attrib(self, graph, expression, attrib_name):
        attrib_value = expression.attrib[attrib_name]
        if "<" in attrib_value and ">" in attrib_value:
            start = attrib_value.find("<")
            end = attrib_value.rfind(">")

            front = attrib_value[:start]
            middle = attrib_value[start:end + 1]
            back = attrib_value[end + 1:]

            root = TemplateNode()
            root.append(TemplateWordNode(front))

            xml = ET.fromstring(middle)
            xml_node = TemplateNode()
            graph.parse_tag_expression(xml, xml_node)
            root.append(xml_node)

            root.append(TemplateWordNode(back))

            self.set_attrib(attrib_name, root)

        else:
            self.set_attrib(attrib_name, TemplateWordNode(attrib_value))

    def _parse_node_with_xml_attribs(self, graph, expression):

        try:
            self._name = TextUtils.tag_from_text(expression.tag)

            for attrib_name in expression.attrib:
                self._parse_attrib(graph, expression, attrib_name)

            self.parse_text(graph, self.get_text_from_element(expression))

            for child in expression:
                graph.parse_tag_expression(child, self)
                self.parse_text(graph, self.get_tail_from_element(child))

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to parse node with attributes", excep)
            raise ParserException("Failed to parse node with attributes")


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.config.container import BaseContainerConfigurationData
from programy.config.bot.bot import BotConfiguration
from programy.scheduling.config import SchedulerConfiguration
from programy.storage.config import StorageConfiguration
from programy.utils.substitutions.substitues import Substitutions
from programy.utils.email.config import EmailConfiguration
from programy.triggers.config import TriggerConfiguration
from programy.clients.ping.config import PingResponderConfig
from programy.config.brain.brain import BrainConfiguration


class ClientConfigurationData(BaseContainerConfigurationData):

    def __init__(self, name):
        BaseContainerConfigurationData.__init__(self, name)
        self._description = 'ProgramY AIML2.0 Client'
        self._renderer = self._get_renderer_class()
        self._scheduler = SchedulerConfiguration()
        self._storage = StorageConfiguration()
        self._email = EmailConfiguration()
        self._triggers = TriggerConfiguration()
        self._responder = PingResponderConfig()

        self._bot_selector = self._get_bot_selector_class()
        bot_config = BotConfiguration('bot')
        self._bot_configs = [bot_config]

    def _get_renderer_class(self):
        return "programy.clients.render.text.TextRenderer"

    def _get_bot_selector_class(self):
        return "programy.clients.botfactory.DefaultBotSelector"

    @property
    def description(self):
        return self._description

    @property
    def configurations(self):
        return self._bot_configs

    @property
    def bot_selector(self):
        return self._bot_selector

    @property
    def scheduler(self):
        return self._scheduler

    @property
    def storage(self):
        return self._storage

    @property
    def renderer(self):
        return self._renderer

    @property
    def email(self):
        return self._email

    @property
    def triggers(self):
        return self._triggers

    @property
    def responder(self):
        return self._responder

    def load_configuration(self, configuration_file, bot_root, subs: Substitutions = None):
        client = configuration_file.get_section(self.section_name)
        if client is None:
            YLogger.error(None, 'Client section named [%s] not found, trying "client"', self.section_name)
            client = configuration_file.get_section('client')

        if client is not None:
            self.load_configuration_section(configuration_file, client, bot_root, subs)

        else:
            YLogger.error(None, 'No client section not found!')

    def load_configuration_section(self, configuration_file, section, bot_root, subs: Substitutions = None):

        assert configuration_file is not None
        assert section is not None

        self._description = configuration_file.get_option(section, "description",
                                                          missing_value='ProgramY AIML2.0 Client', subs=subs)

        self._scheduler.load_config_section(configuration_file, section, bot_root, subs=subs)

        self._storage.load_config_section(configuration_file, section, bot_root, subs=subs)

        self._renderer = configuration_file.get_option(section, "renderer", subs=subs)

        self._email.load_config_section(configuration_file, section, bot_root, subs=subs)

        self._triggers.load_config_section(configuration_file, section, bot_root, subs=subs)

        self._responder.load_config_section(configuration_file, section, bot_root, subs=subs)

        self._load_bot_configurations(configuration_file, section, bot_root, subs)

    def _load_bot_configurations(self, configuration_file, section, bot_root, subs):
        bots = configuration_file.get_section("bots", section)
        if bots is not None:
            bot_keys = configuration_file.get_keys(bots)
            first = True
            for name in bot_keys:
                if first is True:
                    self._bot_configs.clear()
                    first = False
                bot_config = configuration_file.get_section(name, bots)
                config = BotConfiguration(name)
                config.load_configuration_section(configuration_file, bot_config, bot_root, subs=subs)
                self._bot_configs.append(config)

        if len(self._bot_configs) == 0:
            YLogger.warning(self, "No bot name defined for client [%s], defaulting to 'bot'.", self.section_name)
            self._bot_configs.append(BotConfiguration("bot"))
            self._bot_configs[0].configurations.append(BrainConfiguration("brain"))

        self._bot_selector = configuration_file.get_option(section, "bot_selector",
                       missing_value="programy.clients.botfactory.DefaultBotSelector",
                       subs=subs)

    def to_yaml(self, data, defaults=True):

        assert data is not None

        if defaults is True:
            data['description'] = 'ProgramY AIML2.0 Client'
            data['bot_selector'] = "programy.clients.botfactory.DefaultBotSelector"

            data['scheduler'] = {}
            self._scheduler.to_yaml(data['scheduler'], defaults)

            data['email'] = {}
            self._email.to_yaml(data['email'], defaults)

            data['triggers'] = {}
            self._triggers.to_yaml(data['triggers'], defaults)

            data['responder'] = {}
            self._responder.to_yaml(data['responder'], defaults)

            data['renderer'] = "programy.clients.render.text.TextRenderer"

            data['storage'] = {}
            self._storage.to_yaml(data['storage'], defaults)

            data['bots'] = {}
            data['bots']['bot'] = {}
            bot_config = BotConfiguration('bot')
            bot_config.to_yaml(data['bots']['bot'], defaults)

        else:
            data['description'] = self._description
            data['bot_selector'] = self.bot_selector

            data[self._scheduler.id] = {}
            self._scheduler.to_yaml(data[self._scheduler.id], defaults)

            data['email'] = {}
            self._email.to_yaml(data['email'], defaults)

            data['triggers'] = {}
            self._triggers.to_yaml(data['triggers'], defaults)

            data['responder'] = {}
            self._responder.to_yaml(data['responder'], defaults)

            data['renderer'] = self.renderer

            data['storage'] = {}
            self._storage.to_yaml(data['storage'], defaults)

            data['bots'] = {}
            for botconfig in self._bot_configs:
                data['bots'][botconfig.id] = {}
                botconfig.to_yaml(data['bots'][botconfig.id], defaults)




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
from abc import ABC
from abc import abstractmethod
from programy.utils.logging.ylogger import YLogger


class FileFinder(ABC):

    def __init__(self):
        pass    # pragma: no cover

    @abstractmethod
    def load_file_contents(self, fileid, filename, userid="*"):
        raise NotImplementedError()     # pragma: no cover

    def find_files(self, path, subdir=False, extension=None):
        found_files = []

        if subdir is False:
            paths = os.listdir(path)
            for filename in paths:
                if filename.endswith(extension):
                    found_files.append((filename, os.path.join(path, filename)))

        else:
            for dirpath, _, filenames in os.walk(path):
                for filename in [f for f in filenames if f.endswith(extension)]:
                    found_files.append((filename, os.path.join(dirpath, filename)))

        return sorted(found_files, key=lambda element: (element[1], element[0]))

    def load_dir_contents(self, paths, subdir=False, extension=".txt", filename_as_userid=False):

        files = self.find_files(paths, subdir, extension)

        collection = {}
        file_maps = {}
        for file in files:
            just_filename = FileFinder.get_just_filename_from_filepath(file[0])
            try:
                if filename_as_userid is True:
                    userid = just_filename

                else:
                    userid = "*"

                collection[just_filename.upper()] = self.load_file_contents(just_filename, file[1], userid=userid)
                file_maps[just_filename.upper()] = file[1]

            except Exception as excep:
                YLogger.exception(self, "Failed to load file contents for file [%s]", excep, file[1])

        return collection, file_maps

    def load_single_file_contents(self, filename):
        just_filename = FileFinder.get_just_filename_from_filepath(filename)

        collection = {}
        try:
            collection[just_filename] = self.load_file_contents(just_filename, filename)

        except Exception as excep:
            YLogger.exception(self, "Failed to load file contents for file [%s]", excep, filename)

        return collection

    @staticmethod
    def get_just_filename_from_filepath(filepath):

        if os.sep in filepath:
            pathsplits = filepath.split(os.sep)
            filename_ext = pathsplits[-1]
        else:
            filename_ext = filepath

        if "." in filename_ext:
            filesplits = filename_ext.split(".")
            filename = filesplits[0]
        else:
            filename = filename_ext

        return filename



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.utils.logging.ylogger import YLogger
from programy.storage.factory import StorageFactory
from programy.oob.default import DefaultOutOfBandProcessor


class OOBHandler:

    def __init__(self):
        self._oobs = {}

    @property
    def default_oob(self):
        return self._oobs.get('default', None)

    @property
    def oobs(self):
        return self._oobs

    def add_oob(self, name, oob_class):
        self._oobs[name] = oob_class

    def empty(self):
        self._oobs.clear()

    def load_oob_processors(self, storage_factory):
        if storage_factory.entity_storage_engine_available(StorageFactory.OOBS) is True:
            storage_engine = storage_factory.entity_storage_engine(StorageFactory.OOBS)
            oobs_store = storage_engine.oobs_store()
            oobs_store.load(self)
        else:
            YLogger.error(None, "No storage engine available for pattern_nodes!")

        if self._oobs.get('default', None) is None:
            self._oobs['default'] = DefaultOutOfBandProcessor()

    def oob_in_response(self, response):
        if response is not None:
            return "<oob>" in response

    def handle(self, client_context, response):
        if "<oob>" in response:
            response, oob = self.strip_oob(response)
            oob_response = self.process_oob(client_context, oob)
            response = response + " " + oob_response

        return response.strip()

    def strip_oob(self, response):
        match = re.compile(r"(.*)(<\s*oob\s*>.*<\/\s*oob\s*>)(.*)")
        groupings = match.match(response)
        if groupings is not None:
            front = groupings.group(1).strip()
            back = groupings.group(3).strip()
            response = ""
            if front != "":
                response = front + " "

            response += back
            oob = groupings.group(2)
            return response, oob

        return response, None

    def process_oob(self, client_context, oob_command):
        oob_content = ET.fromstring(oob_command)

        if oob_content.tag == 'oob':
            for child in oob_content.findall('./'):
                if child.tag in self._oobs:
                    oob_class = self._oobs[child.tag]
                    return oob_class.process_out_of_bounds(client_context, child)

                def_oob = self.default_oob
                if def_oob is not None:
                    def_oob.process_out_of_bounds(client_context, child)
                else:
                    YLogger.error (client_context, "No default oob defined")

        return ""



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from flask import Flask, request, abort, Response
from linebot import LineBotApi, WebhookParser
from linebot.exceptions import InvalidSignatureError
from linebot.models import MessageEvent, TextMessage, TextSendMessage
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.line.config import LineConfiguration
from programy.utils.console.console import outputLog


class LineBotClient(FlaskRestBotClient):

    def __init__(self, argument_parser=None):
        self._channel_access_token = None
        self._channel_secret = None
        FlaskRestBotClient.__init__(self, "line", argument_parser)

        self.create_line_bot()

        YLogger.debug(self, "Line Client is running....")

    def get_client_configuration(self):
        return LineConfiguration()

    def get_license_keys(self):
        self._channel_secret = self.license_keys.get_key("LINE_CHANNEL_SECRET")
        self._channel_access_token = self.license_keys.get_key("LINE_ACCESS_TOKEN")

    def create_line_bot(self):
        if self._channel_access_token is not None and self._channel_secret is not None:
            self._line_bot_api = LineBotApi(self._channel_access_token)
            self._parser = WebhookParser(self._channel_secret)

        else:
            YLogger.error(self, "Line channel access token and/or secret missing, unable to create line bot")

    def handle_text_message(self, event):
        question = event.message.text
        userid = event.source.user_id

        client_context = self.create_client_context(userid)

        answer = self.ask_question(userid, question)
        rendered = self.renderer.render(client_context, answer)

        self._line_bot_api.reply_message(event.reply_token, TextSendMessage(text=rendered))

    def get_unknown_response(self, userid):
        if self.configuration.client_configuration.unknown_command_srai is None:
            unknown_response = self.configuration.client_configuration.unknown_command

        else:
            unknown_response = self.ask_question(userid, self.configuration.client_configuration.unknown_command_srai)
            if unknown_response is None or unknown_response == "":
                unknown_response = self.configuration.client_configuration.unknown_command

        client_context = self.create_client_context(userid)

        return self.renderer.render(client_context, unknown_response)

    def handle_unknown_event(self, event):
        userid = ""
        unknown_response = self.get_unknown_response(userid)
        self._line_bot_api.reply_message(event.reply_token, TextSendMessage(text=unknown_response))

    def handle_unknown_message(self, event):
        userid = ""
        unknown_response = self.get_unknown_response(userid)
        self._line_bot_api.reply_message(event.reply_token, TextSendMessage(text=unknown_response))

    def handle_message_request(self, body, signature):

        events = self._parser.parse(body, signature)

        for event in events:
            if isinstance(event, MessageEvent):
                if isinstance(event.message, TextMessage):
                    self.handle_text_message(event)

                else:
                    self.handle_unknown_message(event)

            else:
                self.handle_unknown_event(event)

    def receive_message(self, request):

        if self.configuration.client_configuration.debug is True:
            self.dump_request(request)

        # get X-Line-Signature header value
        signature = request.headers['X-Line-Signature']

        # get request body as text
        body = request.get_data(as_text=True)

        # handle webhook body
        try:
            self.handle_message_request(body, signature)
        except InvalidSignatureError as excep:
            YLogger.exception(self, "Line error", excep)
            abort(500)

        return 'OK' #Response(status=200)


if __name__ == "__main__":

    outputLog(None, "Initiating Line Client...")

    LINE_CLIENT = LineBotClient()

    APP = Flask(__name__)

    @APP.route(LINE_CLIENT.configuration.client_configuration.api, methods=['POST'])
    def receive_message():
        try:
            return LINE_CLIENT.receive_message(request)
        except Exception as e:
            YLogger.exception(None, "Line error", e)

    LINE_CLIENT.run(APP)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.dialog.conversation import Conversation
from programy.storage.factory import StorageFactory
from programy.config.bot.conversations import BotConversationsConfiguration


class ConversationManager:

    def __init__(self, conversation_configuration):

        assert conversation_configuration is not None
        assert isinstance(conversation_configuration, BotConversationsConfiguration)

        self._configuration = conversation_configuration
        self._conversation_storage = None
        self._conversations = {}

    @property
    def configuration(self):
        return self._configuration

    @property
    def storage(self):
        return self._conversation_storage

    @property
    def conversations(self):
        return self._conversations

    def empty(self):
        self._conversations.clear()

    def initialise(self, storage_factory):
        if storage_factory.entity_storage_engine_available(StorageFactory.CONVERSATIONS) is True:
            converstion_engine = storage_factory.entity_storage_engine(StorageFactory.CONVERSATIONS)
            if converstion_engine:
                self._conversation_storage = converstion_engine.conversation_store()

    def save_conversation(self, client_context):
        if self._conversation_storage is not None:
            if client_context.userid in self._conversations:
                conversation = self._conversations[client_context.userid]
                self._conversation_storage.store_conversation(client_context, conversation)

    def has_conversation(self, client_context) -> bool:
        return bool(client_context.userid in self._conversations)

    def get_conversation(self, client_context):

        assert client_context is not None
        assert client_context.userid is not None

        if client_context.userid in self._conversations:
            YLogger.debug(client_context, "Retrieving conversation for client %s", client_context.userid)
            conversation = self._conversations[client_context.userid]

            # Load existing conversation from cache
            if self.configuration.multi_client:
                if self._conversation_storage is not None:
                    self._conversation_storage.load_conversation(client_context, conversation)

        else:
            YLogger.info(client_context, "Creating new conversation for client %s", client_context.userid)

            conversation = Conversation(client_context)

            if client_context.brain.default_variables is not None:
                conversation.load_initial_variables(client_context.brain.default_variables)

            self._conversations[client_context.userid] = conversation

            if self._conversation_storage is not None:
                self._conversation_storage.load_conversation(client_context, conversation)

            if self.configuration.restore_last_topic is True:
                pass

        return conversation



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
import re
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.utils.logging.ylogger import YLogger
from programy.parser.exceptions import ParserException
from programy.parser.exceptions import DuplicateGrammarException
from programy.parser.pattern.graph import PatternGraph
from programy.parser.template.graph import TemplateGraph
from programy.utils.files.filefinder import FileFinder
from programy.dialog.sentence import Sentence
from programy.parser.pattern.matchcontext import MatchContext
from programy.storage.factory import StorageFactory
from programy.utils.console.console import outputLog
from programy.utils.text.text import TextUtils


class AIMLLoader(FileFinder):

    def __init__(self, aiml_parser):
        FileFinder.__init__(self)
        self._aiml_parser = aiml_parser

    @property
    def parser(self):
        return self._aiml_parser

    def load_file_contents(self, fileid, filename, userid="*"):
        del fileid
        return self._aiml_parser.parse_from_file(filename, userid=userid)


class AIMLParser:
    RE_PATTERN_OF_TAG_AND_NAMESPACE_FROM_TEXT = re.compile(r"^\{.*\}.*$")
    RE_MATCH_OF_TAG_AND_NAMESPACE_FROM_TEXT = re.compile(r"^(\{.*\})(.*)$")

    def __init__(self, brain):
        self._brain = brain
        self._pattern_parser = self.create_pattern_graph()
        self._template_parser = self.create_template_graph()
        self._aiml_loader = self.create_aiml_loader()
        self._num_categories = 0
        self._duplicates = None
        self._errors = None

    def create_pattern_graph(self):
        return PatternGraph(aiml_parser=self)

    def create_template_graph(self):
        return TemplateGraph(aiml_parser=self)

    def create_aiml_loader(self):
        return AIMLLoader(self)

    def __getstate__(self):
        # We don't need to pickle the File Writes for duplicates and errors,
        # __getstate__ is called during the pickling process to determin whih
        # attributes to load, so we remove the ones we don't want pickling
        d = dict(self.__dict__)
        del d['_brain']
        if '_errors' in d:
            del d['_errors']
        if '_duplicates' in d:
            del d['_duplicates']
        return d

    @property
    def brain(self):
        return self._brain

    @property
    def num_categories(self):
        return self._num_categories

    @property
    def pattern_parser(self):
        return self._pattern_parser

    @property
    def template_parser(self):
        return self._template_parser

    def empty(self):
        self._pattern_parser.empty()

    def load_aiml(self):

        self.create_debug_storage()

        if self.brain.bot.client.storage_factory.entity_storage_engine_available(StorageFactory.CATEGORIES) is True:
            storage_engine = self.brain.bot.client.storage_factory.entity_storage_engine(StorageFactory.CATEGORIES)
            category_store = storage_engine.category_store()
            category_store.load_all(self)
        else:
            YLogger.error(None, "No category storage defined, no aiml loaded!")

        self.save_debug_files()
        self.display_debug_info()

    @staticmethod
    def check_aiml_tag(aiml, filename=None):
        # Null check just to be sure
        if aiml is None:
            raise ParserException("Null root tag", filename=filename)

        tag_name, namespace = TextUtils.tag_and_namespace_from_text(aiml.tag)

        # Then if check is just <aiml>, thats OK
        if tag_name != 'aiml':
            raise ParserException("Root tag is not <aiml>", filename=filename)

        return tag_name, namespace

    def parse_from_file(self, filename, userid="*"):
        YLogger.info(self, "Loading aiml file: " + filename)

        try:
            tree = ET.parse(filename, parser=LineNumberingParser())
            aiml = tree.getroot()

            _, namespace = AIMLParser.check_aiml_tag(aiml, filename=filename)

            start = datetime.datetime.now()
            num_categories = self.parse_aiml(aiml, namespace, filename, userid=userid)
            stop = datetime.datetime.now()
            diff = stop - start
            YLogger.info(self, "Processed %s with %d categories in %f.2 secs", filename, num_categories,
                         diff.total_seconds())

        except Exception as excep:
            YLogger.exception(self, "Failed to load contents of AIML file from [%s]", excep, filename)

    def parse_from_text(self, text):
        """
         Parse an AIML text version of an aiml file and return all the cateogeries found in the file
         :param text: Fully validated AIML snippet
         :return list of categories parsed from file:
         """

        aiml = ET.fromstring(text)

        _, namespace = AIMLParser.check_aiml_tag(aiml)

        self.parse_aiml(aiml, namespace)

    def create_debug_storage(self):
        if self.brain.configuration.debugfiles.save_errors is True:
            self._errors = []
        if self.brain.configuration.debugfiles.save_duplicates is True:
            self._duplicates = []

    def save_debug_files(self):
        if self.brain.configuration.debugfiles.save_errors is True:
            if self.brain.bot.client.storage_factory.entity_storage_engine_available(StorageFactory.ERRORS) is True:
                storage_engine = self.brain.bot.client.storage_factory.entity_storage_engine(StorageFactory.ERRORS)
                errors_store = storage_engine.errors_store()
                errors_store.save_errors(self._errors)

        if self.brain.configuration.debugfiles.save_duplicates is True:
            if self.brain.bot.client.storage_factory.entity_storage_engine_available(StorageFactory.DUPLICATES) is True:
                storage_engine = self.brain.bot.client.storage_factory.entity_storage_engine(StorageFactory.DUPLICATES)
                duplicates_store = storage_engine.duplicates_store()
                duplicates_store.save_duplicates(self._duplicates)

    def display_debug_info(self):
        if self._errors is not None:
            outputLog(self, "Found a total of %d errors in your grammars, check your errors store" % len(self._errors))
        if self._duplicates is not None:
            outputLog(self, "Found a total of %d duplicates in your grammars, check your duplicates store" % len(
                self._duplicates))

    def handle_aiml_duplicate(self, dupe_excep, filename, expression):
        if self._duplicates is not None:
            dupe_excep.filename = filename
            msg = dupe_excep.format_message()
            YLogger.error(self, msg)

            startline = None
            endline = None
            if expression is not None:
                if hasattr(expression, "_start_line_number"):
                    startline = str(expression._start_line_number)  # pylint: disable=protected-access

                if hasattr(expression, "_end_line_number"):
                    endline = str(expression._end_line_number)  # pylint: disable=protected-access

            self._duplicates.append([dupe_excep.message, filename, startline, endline])

    def handle_aiml_error(self, parser_excep, filename, expression):
        if self._errors is not None:
            parser_excep.filename = filename
            msg = parser_excep.format_message()
            YLogger.error(self, msg)

            startline = None
            endline = None
            if expression is not None:
                if hasattr(expression, "_start_line_number"):
                    startline = str(expression._start_line_number)  # pylint: disable=protected-access

                if hasattr(expression, "_end_line_number"):
                    endline = str(expression._end_line_number)  # pylint: disable=protected-access

            self._errors.append([parser_excep.message, filename, startline, endline])

    def parse_aiml(self, aiml_xml, namespace, filename=None, userid="*"):
        del namespace

        self.parse_version(aiml_xml)

        categories_found = False
        num_category = 0
        for expression in aiml_xml:
            tag_name, namespace = TextUtils.tag_and_namespace_from_text(expression.tag)
            if tag_name == 'topic':
                try:
                    num_topic_categories = self.parse_topic(expression, namespace)
                    num_category += num_topic_categories
                    categories_found = True

                except DuplicateGrammarException as dupe_excep:
                    self.handle_aiml_duplicate(dupe_excep, filename, expression)

                except ParserException as parser_excep:
                    self.handle_aiml_error(parser_excep, filename, expression)

            elif tag_name == 'category':
                try:
                    self.parse_category(expression, namespace, userid=userid)
                    categories_found = True
                    num_category += 1

                except DuplicateGrammarException as dupe_excep:
                    self.handle_aiml_duplicate(dupe_excep, filename, expression)

                except ParserException as parser_excep:
                    self.handle_aiml_error(parser_excep, filename, expression)

            else:
                raise ParserException("Unknown top level tag, %s" % expression.tag, xml_element=expression)

        if categories_found is False:
            if filename is not None:
                YLogger.warning(self, "no categories in aiml file")
            else:
                YLogger.warning(self, "no categories in aiml file [%s]"% filename)

        return num_category

    def parse_version(self, aiml):
        if 'version' in aiml.attrib:
            version = aiml.attrib['version']
            if version not in ['0.9', '1.0', '1.1', '2.0']:
                YLogger.warning(self, "Version number not a supported version: %s", version)
        else:
            YLogger.warning(self, "No version info, defaulting to 2.0")
            version = "2.0"
        return version

    def parse_topic(self, topic_element, namespace):

        if 'name' in topic_element.attrib:
            name = topic_element.attrib['name']
            if name is None or not name:
                raise ParserException("Topic name empty or null", xml_element=topic_element)
            xml = "<topic>%s</topic>" % name
            YLogger.info(self, "Topic attrib converted to %s", xml)
            topic_pattern = ET.fromstring(xml)
        else:
            raise ParserException("Missing name attribute for topic", xml_element=topic_element)

        category_found = False
        num_category = 0
        for child in topic_element:
            tag_name, _ = TextUtils.tag_and_namespace_from_text(child.tag)
            if tag_name == 'category':
                self.parse_category(child, namespace, topic_pattern)
                category_found = True
                num_category += 1
            else:
                raise ParserException("Unknown child node of topic, %s" % child.tag, xml_element=topic_element)

        if category_found is False:
            raise ParserException("No categories in topic", xml_element=topic_element)

        return num_category

    def find_all(self, element, name, namespace):
        if namespace is not None:
            search = '%s%s' % (namespace, name)
            return element.findall(search)
        return element.findall(name)

    def find_topic(self, category_xml, namespace, topic_element=None):
        topics = self.find_all(category_xml, "topic", namespace)

        if topic_element is not None:
            if topics:
                raise ParserException("Topic exists in category AND as parent node", xml_element=category_xml)

        else:
            if len(topics) > 1:
                raise ParserException("Multiple <topic> nodes found in category", xml_element=category_xml)
            elif len(topics) == 1:
                topic_element = topics[0]
            else:
                topic_element = ET.fromstring("<topic>*</topic>")

        return topic_element

    def find_that(self, category_xml, namespace):
        thats = self.find_all(category_xml, "that", namespace)
        if len(thats) > 1:
            raise ParserException("Multiple <that> nodes found in category", xml_element=category_xml)
        elif len(thats) == 1:
            that_element = thats[0]
        else:
            that_element = ET.fromstring("<that>*</that>")
        return that_element

    def get_template(self, category_xml, namespace):
        templates = self.find_all(category_xml, "template", namespace)
        if not templates:
            raise ParserException("No template node found in category", xml_element=category_xml)
        elif len(templates) > 1:
            raise ParserException("Multiple <template> nodes found in category", xml_element=category_xml)
        else:
            return self._template_parser.parse_template_expression(templates[0])

    def get_pattern(self, category_xml, namespace):
        patterns = self.find_all(category_xml, "pattern", namespace)
        if not patterns:
            raise ParserException("No pattern node found in category", xml_element=category_xml)
        elif len(patterns) > 1:
            raise ParserException("Multiple <pattern> nodes found in category", xml_element=category_xml)
        else:
            return patterns[0]

    def parse_category(self, category_xml, namespace, topic_element=None, add_to_graph=True, userid="*"):

        topic_element = self.find_topic(category_xml, namespace, topic_element)

        that_element = self.find_that(category_xml, namespace)

        template_graph_root = self.get_template(category_xml, namespace)

        pattern = self.get_pattern(category_xml, namespace)

        if add_to_graph is True:
            self._pattern_parser.add_pattern_to_graph(pattern, topic_element, that_element, template_graph_root,
                                                      userid=userid)
            self._num_categories += 1

        return (pattern, topic_element, that_element, template_graph_root)

    def match_sentence(self, client_context, pattern_sentence, topic_pattern, that_pattern):

        topic_sentence = Sentence(client_context, topic_pattern)
        that_sentence = Sentence(client_context, that_pattern)

        YLogger.debug(client_context, "AIML Parser matching sentence [%s], topic=[%s], that=[%s] ",
                      pattern_sentence.text(client_context), topic_pattern, that_pattern)

        sentence = Sentence(client_context)
        sentence.append_sentence(pattern_sentence)
        sentence.append_word('__TOPIC__')
        sentence.append_sentence(topic_sentence)
        sentence.append_word('__THAT__')
        sentence.append_sentence(that_sentence)
        YLogger.debug(client_context, "Matching [%s]", sentence.words_from_current_pos(client_context, 0))

        context = MatchContext(max_search_depth=client_context.bot.configuration.max_search_depth,
                               max_search_timeout=client_context.bot.configuration.max_search_timeout)

        template = self._pattern_parser.root.match(client_context, context, sentence)

        if template is not None:
            context.template_node = template

            context.list_matches(client_context)

            # Save the matched context for the associated sentence
            pattern_sentence.matched_context = context

            return context

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import time
import datetime
import tweepy
from tweepy.error import RateLimitError
from programy.utils.logging.ylogger import YLogger
from programy.clients.polling.client import PollingBotClient
from programy.clients.polling.twitter.config import TwitterConfiguration
from programy.utils.console.console import outputLog


class TwitterListener(tweepy.StreamListener):

    def __init__(self, twitter_bot_client):
        self._twitter_bot_client = twitter_bot_client
        self.api = twitter_app._api

    def on_status(self, status):
        self._twitter_bot_client.handle_mention(status)

    def on_error(self, status):
        YLogger.error(self, status)


class TwitterBotClient(PollingBotClient):
    FIFTEEN_MINUTES = 15 * 60

    def __init__(self, argument_parser=None):
        self._username = "unknown"
        self._username_len = 0
        self._welcome_message = None
        self._api = None
        self._me = None
        self._consumer_key = None
        self._consumer_secret = None
        self._access_token = None
        self._access_token_secret = None

        self._direct_id = -1

        PollingBotClient.__init__(self, "twitter", argument_parser)

    def get_client_configuration(self):
        return TwitterConfiguration()

    def get_license_keys(self):
        self._username = self.license_keys.get_key("TWITTER_USERNAME")
        self._consumer_key = self.license_keys.get_key("TWITTER_CONSUMER_KEY")
        self._consumer_secret = self.license_keys.get_key("TWITTER_CONSUMER_SECRET")
        self._access_token = self.license_keys.get_key("TWITTER_ACCESS_TOKEN")
        self._access_token_secret = self.license_keys.get_key("TWITTER_ACCESS_TOKEN_SECRET")
        self._username_len = len(self._username)  # Going to get used quite a lot

    def _create_api(self, consumer_key, consumer_secret, access_token, access_token_secret):
        auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
        auth.set_access_token(access_token, access_token_secret)
        api = tweepy.API(auth, wait_on_rate_limit=False, wait_on_rate_limit_notify=False)
        try:
            api.verify_credentials()
            return api

        except Exception as error:
            YLogger.exception(self, "Error creating api", error)
            return None

    def connect(self):
        if self._consumer_key is not None and \
                self._consumer_secret is not None and \
                self._access_token is not None and \
                self._access_token_secret is not None:
            self._welcome_message = self.configuration.client_configuration.welcome_message
            self._api = self._create_api(self._consumer_key,
                                         self._consumer_secret,
                                         self._access_token,
                                         self._access_token_secret)
            self._me = self._api.me()
            return True

        return False

    def display_connected_message(self):
        outputLog(self, "Twitter Bot connected and running...")

    def ask_question(self, userid, question):
        self._questions += 1
        client_context = self.create_client_context(userid)
        return client_context.bot.ask_question(client_context, question, responselogger=self)

    def _follow_followers(self):
        YLogger.debug(self, "Retrieving and following followers")
        for follower in tweepy.Cursor(self._api.followers).items():
            if not follower.following:
                YLogger.debug(self, f"Following {follower.name}")
                follower.follow()

    def handle_mention(self, tweet):
        if tweet.in_reply_to_status_id is not None or tweet.user.id == self._me.id:
            # This tweet is a reply or I'm its author so, ignore it
            return

        if not tweet.favorited:
            # Mark it as Liked, since we have not done it yet
            try:
                tweet.favorite()

            except Exception as e:
                YLogger.error("Error on fav", exc_info=True)

        # Reply to the user
        try:
            author = tweet.author.screen_name

            question = " ".join(tweet.text.split(" ")[1:])
            YLogger.debug(self, "Mention: %s", question)

            response = self.ask_question(author, question)

            reply = "@{0} {1}".format(author, response)
            YLogger.debug(self, "Mention reply: %s", reply)

            self._api.update_status(reply, in_reply_to_status_id=tweet.id)

        except Exception as error:
            YLogger.exception(self, "Error handling mention", error)

    def _check_directs(self):
        try:
            YLogger.info(self, "Retrieving directs")
            new_since_id = self._direct_id
            if self._direct_id == -1:
                tweets = tweepy.Cursor(self._api.list_direct_messages).items()
            else:
                tweepy.Cursor(self._api.list_direct_messages, count=10, cursor=new_since_id).items()

            for tweet in tweets:
                new_since_id = max(tweet.id, new_since_id)

            self._direct_id = new_since_id

        except Exception as error:
            YLogger.exception(self, "Failure handling direct", error)

    def pre_poll(self):
        if self.configuration.client_configuration.respond_to_mentions is True:
            YLogger.debug(self, "Responding to mentions: %s", self.configuration.client_configuration.mentions)
            stream = tweepy.Stream(self._api.auth, TwitterListener(self))
            stream.filter(track=self.configuration.client_configuration.mentions, languages=["en"], is_async=True)

    def poll_and_answer(self):

        running = True
        sleepy_time = self._configuration.client_configuration.polling_interval

        try:
            if self._configuration.client_configuration.follow_followers is True:
                YLogger.debug(self, "Following followers")
                self._follow_followers()

            if self._configuration.client_configuration.respond_to_directs:
                YLogger.debug(self, "Checking direct messages")
                self._check_directs()

            YLogger.debug(self, "Sleeping for %d at %s" %(sleepy_time, datetime.datetime.now().strftime("%H:%M:%S")))

        except KeyboardInterrupt:
            running = False

        except RateLimitError:

            outputLog(self, "Rate limit exceeded, check logs!")
            YLogger.error(self, "Rate limit exceeded, sleeping for %d seconds", self._rate_limit_sleep)

            sleepy_time = self._rate_limit_sleep

        except Exception as excep:
            YLogger.exception(self, "Poll and answer error", excep)

        time.sleep(sleepy_time)

        return running


if __name__ == '__main__':
    outputLog(None, "Initiating Twitter Client...")

    twitter_app = TwitterBotClient()
    twitter_app.run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.parser.template.nodes.eval import TemplateEvalNode
from programy.utils.text.text import TextUtils
from programy.parser.exceptions import ParserException


class LearnCategory:

    def __init__(self, pattern, topic, that, template):
        self._pattern = pattern
        self._topic = topic
        self._that = that
        self._template = template
        self._children = []

    @property
    def pattern(self):
        return self._pattern

    @pattern.setter
    def pattern(self, pattern):
        self._pattern = pattern

    @property
    def topic(self):
        return self._topic

    @topic.setter
    def topic(self, topic):
        self._topic = topic

    @property
    def that(self):
        return self._that

    @that.setter
    def that(self, that):
        self._that = that

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, template):
        self._template = template

    @property
    def children(self):
        return self._children

    def append(self, category):
        self._children.append(category)

    def to_string(self):
        return "[CATEGORY]"


class TemplateLearnNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)

    def evaluate_eval_nodes(self, client_context, template):

        new_template = client_context.brain.aiml_parser.template_parser.get_base_node()

        count = 0
        for child in template.children:
            if isinstance(child, TemplateEvalNode):
                new_word_node = client_context.brain.aiml_parser.template_parser.get_word_node(
                    child.resolve(client_context))
                new_template.children.append(new_word_node)
            else:
                new_template.children.append(child)
            count += 1

        return new_template

    def resolve_element_evals(self, client_context, element):

        new_element = ET.Element(element.tag)

        new_element.text = TextUtils.strip_whitespace(element.text)

        for child in element:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'eval':
                eval_str = ET.tostring(child, 'utf-8').decode('ascii')
                eval_str = TextUtils.strip_whitespace(eval_str)
                str_val = "<template>%s</template>" % eval_str
                template = ET.fromstring(str_val)

                ast = client_context.brain.aiml_parser.template_parser.parse_template_expression(template)
                resolved = ast.resolve(client_context)

                new_element.text += " " + resolved
            else:
                new_element.append(child)

        new_element.text = new_element.text.upper()

        if element.tail is not None:
            new_element.tail = TextUtils.strip_whitespace(element.tail)

        return new_element

    def _create_new_category(self, client_context, category, userid="*"):
        del userid

        new_pattern = self.resolve_element_evals(client_context, category.pattern)
        new_topic = self.resolve_element_evals(client_context, category.topic)
        new_that = self.resolve_element_evals(client_context, category.that)

        new_template = self.evaluate_eval_nodes(client_context, category.template)

        client_context.brain.aiml_parser.pattern_parser.add_pattern_to_graph(new_pattern, new_topic, new_that,
                                                                             new_template, learn=True,
                                                                             userid=client_context.userid)

        YLogger.debug(client_context, "[%s] resolved to new pattern [[%s] [%s] [%s]", self.to_string(),
                      ET.tostring(new_pattern, 'utf-8').decode('utf-8'),
                      ET.tostring(new_topic, 'utf-8').decode('utf-8'),
                      ET.tostring(new_that, 'utf-8').decode('utf-8'))

        return LearnCategory(new_pattern, new_topic, new_that, new_template)

    def resolve_to_string(self, client_context):
        for category in self.children:
            self._create_new_category(client_context, category, userid=client_context.userid)
        return ""

    def to_string(self):
        return "[LEARN]"

    def children_to_xml(self, client_context):
        xml = ""
        for category in self.children:
            xml += "<category>"
            xml += ET.tostring(category.pattern, 'utf-8').decode('utf-8')
            xml += ET.tostring(category.topic, 'utf-8').decode('utf-8')
            xml += ET.tostring(category.that, 'utf-8').decode('utf-8')
            xml += "<template>"
            xml += category.template.to_xml(client_context)
            xml += "</template>"
            xml += "</category>"
        return xml

    def to_xml(self, client_context):
        xml = "<learn>"
        xml += self.children_to_xml(client_context)
        xml += "</learn>"
        return xml

    def parse_expression(self, graph, expression):

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'category':
                parsed = graph.aiml_parser.parse_category(child, namespace=None, topic_element=None, add_to_graph=False)
                learn_category = LearnCategory(parsed[0], parsed[1], parsed[2], parsed[3])
                self.children.append(learn_category)

            elif tag_name == 'topic':
                raise ParserException("Not supported yet")

            else:
                raise ParserException("Invalid tag [%s] found in <learn>" % tag_name)
